import torch
from torch import nn as nn


biases = False


class Pool2BN(nn.Module):
    def __init__(self, num_channels):
        super().__init__()
        self.bn = torch.nn.BatchNorm1d(num_channels * 2)

    def forward(self, x):
        avgp = torch.nn.functional.adaptive_avg_pool1d(x, 1)[:, :, 0]
        maxp = torch.nn.functional.adaptive_max_pool1d(x, 1)[:, :, 0]
        x = torch.cat((avgp, maxp), axis=1)
        x = self.bn(x)
        return x


class SimpleCNN(torch.nn.Module):
    def __init__(self, k, num_filters, sigmoid=False, additional_layer=False):
        super(SimpleCNN, self).__init__()
        self.sigmoid = sigmoid
        self.cnn = torch.nn.Conv1d(in_channels=4, out_channels=num_filters, kernel_size=k, bias=biases)

        self.additional_layer = additional_layer
        if additional_layer:
            self.bn = nn.BatchNorm1d(num_filters)
            # self.do = nn.Dropout(0.5)
            self.cnn2 = nn.Conv1d(in_channels=num_filters, out_channels=num_filters, kernel_size=1, bias=biases)

        self.post = Pool2BN(num_filters)

    def forward(self, x):
        x = self.cnn(x)
        x = (torch.tanh if self.sigmoid else torch.relu)(x)

        if self.additional_layer:
            x = self.bn(x)
            # x = self.do(x)
            x = self.cnn2(x)
            x = (torch.tanh if self.sigmoid else torch.relu)(x)

        x = self.post(x)
        return x


class MLP(torch.nn.Module):
    def __init__(self, layer_sizes, biases=False, sigmoid=False, dropout=None):
        super().__init__()
        layers = []
        prev_size = layer_sizes[0]
        for i, s in enumerate(layer_sizes[1:]):
            if i != 0 and dropout is not None:
                layers.append(torch.nn.Dropout(dropout))

            layers.append(torch.nn.Linear(in_features=prev_size, out_features=s, bias=biases))
            if i != len(layer_sizes) - 2:
                if sigmoid:
                    # layers.append(torch.nn.Sigmoid())
                    layers.append(torch.nn.Tanh())
                else:
                    layers.append(torch.nn.ReLU())

                layers.append(torch.nn.BatchNorm1d(s))

            prev_size = s

        self.mlp = torch.nn.Sequential(*layers)

    def forward(self, x):
        return self.mlp(x)


class JoinNet(torch.nn.Module):
    def __init__(self, net1, net2, head):
        super().__init__()
        self.net1 = net1
        self.net2 = net2
        self.head = head

    def forward(self, x1, x2):
        x1 = self.net1(x1)
        x2 = self.net2(x2)
        joined = torch.cat((x1, x2), axis=1)
        return self.head(joined)


class ResNet1dBlock(torch.nn.Module):
    def __init__(self, num_filters, k1, internal_filters, k2, dropout=None, dilation=None):
        super().__init__()

        self.init_do = torch.nn.Dropout(dropout) if dropout is not None else None
        self.bn1 = torch.nn.BatchNorm1d(num_filters)
        if dilation is None:
            dilation = 1

        self.cnn1 = torch.nn.Conv1d(in_channels=num_filters, out_channels=internal_filters, kernel_size=k1, bias=biases,
                                    dilation=dilation,
                                    padding=(k1 // 2) * dilation)

        self.bn2 = torch.nn.BatchNorm1d(internal_filters)
        self.cnn2 = torch.nn.Conv1d(in_channels=internal_filters, out_channels=num_filters, kernel_size=k2, bias=biases,
                                    padding=k2 // 2)

    def forward(self, x):
        x_orig = x

        x = self.bn1(x)
        x = torch.relu(x)
        if self.init_do is not None:
            x = self.init_do(x)

        x = self.cnn1(x)

        x = self.bn2(x)
        x = torch.relu(x)
        x = self.cnn2(x)

        return x + x_orig


class ResNet1d(torch.nn.Module):
    def __init__(self, num_filters, block_spec, dropout=None, dilation=None):
        super().__init__()

        blocks = [ResNet1dBlock(num_filters, *spec, dropout=dropout, dilation=dilation) for spec in block_spec]
        self.blocks = torch.nn.Sequential(*blocks)

    def forward(self, x):
        return self.blocks(x)


class NTXentLoss(torch.nn.Module):
    def __init__(self, temperature, use_cosine_similarity):
        super(NTXentLoss, self).__init__()
        self.temperature = temperature
        self.similarity_function = self._get_similarity_function(use_cosine_similarity)

    def _get_similarity_function(self, use_cosine_similarity):
        if use_cosine_similarity:
            self._cosine_similarity = torch.nn.CosineSimilarity(dim=-1)
            return self._cosine_simililarity
        else:
            return self._dot_simililarity

    @staticmethod
    def _dot_simililarity(x, y):
        v = torch.tensordot(x.unsqueeze(1), y.T.unsqueeze(0), dims=2)
        # x shape: (N, 1, C)
        # y shape: (1, C, 2N)
        # v shape: (N, 2N)
        return v

    def _cosine_simililarity(self, x, y):
        # x shape: (N, 1, C)
        # y shape: (1, 2N, C)
        # v shape: (N, 2N)
        v = self._cosine_similarity(x.unsqueeze(1), y.unsqueeze(0))
        return v

    def forward(self, zs, labs):
        batch_size = len(labs)
        similarity_matrix = self.similarity_function(zs, zs) / self.temperature
        labs_row = labs.unsqueeze(0).expand(batch_size, -1)
        same_lab = labs_row == labs_row.transpose(0, 1)
        same_lab.fill_diagonal_(0)
        pos_sims = similarity_matrix[same_lab]
        denoms = torch.log(torch.exp(similarity_matrix).sum(axis=1) - torch.exp(similarity_matrix.diag())).unsqueeze(1).expand(batch_size, batch_size)
        pos_denoms = denoms[same_lab]
        n_pos = len(pos_sims)
        loss = -pos_sims + pos_denoms
        return loss.sum() / n_pos
