#pragma once

#include "misc.h"
#include <fstream>

using namespace std;


const int num_labs = 1314;


typedef uint64_t hash_t;


char complement(char n) {
	switch(n) {
	case 'A':
		return 'T';
	case 'C':
		return 'G';
	case 'G':
		return 'C';
	case 'T':
		return 'A';
	case 'N':
		return 'N';
	}

	return 'X';
}


string reverse_complement(const string& dna) {
	string rc;
	rc.reserve(dna.size());
	for(auto it = dna.rbegin(); it != dna.rend(); ++it)
		rc.push_back(complement(*it));

	return rc;
}



template<class T>
void serialize_value(istream &file, T& value, int64_t count = 1) {
	file.read((char*) &value, sizeof(T) * count);
}


template<class T>
void serialize_value(ostream &file, const T& value, int64_t count = 1) {
	file.write((char*) &value, sizeof(T) * count);
}


template<class T>
void serialize_sized(ostream& destination, const T& v) {
	uint64_t size = v.size();
	serialize_value(destination, size);
	if (size > 0)
		serialize_value(destination, v[0], v.size());
}


template<class T>
void serialize_sized(istream& source, T& v) {
	uint64_t size;
	serialize_value(source, size);
	if(!source)
		return;

	v.resize((size_t)size);
	if (size > 0)
		serialize_value(source, v[0], v.size());
}


class random_hasher {
	vector<vector<hash_t>> letter_hashes, letter_hashes_complement;
	hash_t hbegin, hend;
public:
	random_hasher(int seed, int max_size) : letter_hashes(max_size, vector<hash_t>(256)), letter_hashes_complement(max_size, vector<hash_t>(256)) {
	    mt19937_64 g(seed);
	    for(int p = 0; p < max_size; ++p)
	    	for(char n : {'A', 'C', 'G', 'T', 'N'}) {
	    		letter_hashes[p][n] = g();
	    		letter_hashes_complement[p][complement(n)] = letter_hashes[p][n];
	    	}

	    hbegin = g();
	    hend = g();
	}

	random_hasher(istream& s) {
		int max_size;
		serialize_value(s, max_size);
		letter_hashes.resize(max_size);
		letter_hashes_complement.resize(max_size);
		for(int i = 0; i < max_size; ++i) {
			serialize_sized(s, letter_hashes[i]);
			serialize_sized(s, letter_hashes_complement[i]);
		}

		serialize_value(s, hbegin);
		serialize_value(s, hend);
	}

	template<class Func>
	void for_all_kmers(const string& s, const vector<int>& ks, Func func) const {
		if(ks.front() > (int)s.size())
			return;

		for(int p = 0; p < (int)s.size() - ks.front() + 1; ++p) {
			for(int size : ks) {
				if(p + size > (int)s.size())
					break;

				hash_t h = substr_hash(s, p, size);
				if(h != 0) {
					func(h, p, size);
					if(p == 0 || p + size == (int)s.size())
						func(h ^ hbegin, p, size);

					//if(p + size == (int)s.size())
					//	func(h ^ hend, p, size);
				}
			}
		}
	}

	hash_t substr_hash(const string& s, int p, int size) const {
		hash_t hf = 0, hrc = 0;
		for(int np = 0; np < size; ++np) {
			char l = s[p + np];
			//if(l == 'N') {
			//	hf = 0;
			//	break;
			//}

			hf ^= letter_hashes[np][l];
			hrc ^= letter_hashes_complement[size - np - 1][l];
		}

		return min(hf, hrc);
	}

	void save(ostream& s) const {
		int max_size = letter_hashes.size();
		serialize_value(s, max_size);
		for(int i = 0; i < max_size; ++i) {
			serialize_sized(s, letter_hashes[i]);
			serialize_sized(s, letter_hashes_complement[i]);
		}

		serialize_value(s, hbegin);
		serialize_value(s, hend);
	}
};


struct hash_pair {
    template <class T1, class T2>
    size_t operator()(const pair<T1, T2>& p) const
    {
        auto hash1 = hash<T1>{}(p.first);
        auto hash2 = hash<T2>{}(p.second);
        return hash1 ^ hash2;
    }
};


double box_cox_transform(double x, double lambda=-1.943148224586308) {
	return (pow(x, lambda) - 1) / lambda;
}


class hash_lookup {
	struct count_entry {
		uint32_t lab;
		double weight;
	};

	struct hash_entry {
		int32_t idx;
		atomic<uint32_t> num_labs;

		void operator=(const hash_entry& e) {
			idx = e.idx;
			num_labs = (int)e.num_labs;
		}
	};

	vector<count_entry> counts;
	unordered_map<hash_t, hash_entry> table;
	vector<int> ks;
	random_hasher hasher;

	vector<double> multipliers;
public:
	hash_lookup(const vector<string>& strings, const vector<uint32_t>& labs, const vector<int>& ks, int max_labs_for_kmer, bool verbose) : ks(ks),
	hasher(2, ks.back()) {
		vector<unordered_map<hash_t, uint32_t>> tables(num_labs);

		unique_ptr<progresscpp::ProgressBar> progress_bar;
		if(verbose)
			progress_bar.reset(new progresscpp::ProgressBar(strings.size(), 100, "Collecting hashes from " + to_string(strings.size()) + " sequences"));

		mutex table_mutex[num_labs];

#pragma omp parallel for
		for(int i = 0; i < (int)strings.size(); ++i) {
			unordered_set<hash_t> seen_hashes;
			hasher.for_all_kmers(strings[i], ks, [this, &labs, i, &tables, &seen_hashes, &table_mutex](hash_t h, int p, int size) {
				if(seen_hashes.count(h) != 0)
					return;

				seen_hashes.insert(h);
				lock_guard<mutex> guard(table_mutex[labs[i]]);
				tables[labs[i]][h]++;
			});

			if(verbose)
				++*progress_bar;
		}

		size_t num_hashes_before_join = 0;
		for(auto& t : tables)
			num_hashes_before_join += t.size();

		if(verbose) {
			progress_bar->done();
			progress_bar.reset(new progresscpp::ProgressBar(tables.size(), 100, "Merging " + to_string(num_hashes_before_join) + " hashes from all " + to_string(num_labs) + " labs"));
		}

		unordered_map<hash_t, uint32_t> hcounts;
		size_t num_hashes_removed = 0;
		for(auto& t : tables) {
			for(auto& kvp : t)
				if(++hcounts[kvp.first] == max_labs_for_kmer + 1)
					++num_hashes_removed;

			if(verbose)
				++*progress_bar;
		}

		if(verbose) {
			progress_bar->done();
			progress_bar.reset(new progresscpp::ProgressBar(num_labs, 100, "Removing hashes present in more than " + to_string(max_labs_for_kmer) + " labs"));
		}

#pragma omp parallel for
		for(int l = 0; l < num_labs; ++l) {
			for(auto it = tables[l].begin(); it != tables[l].end();)
				if(hcounts[it->first] > max_labs_for_kmer)
					it = tables[l].erase(it);
				else
					++it;

			if(verbose)
				++*progress_bar;
		}

		size_t total_lab_counts = 0;
		for(auto& t : tables)
			total_lab_counts += t.size();

		if(verbose) {
			progress_bar->done();
			progress_bar.reset(new progresscpp::ProgressBar(hcounts.size(), 100, "Preparing compact index layout"));
		}

		counts.resize(total_lab_counts);
		table.reserve(hcounts.size() - num_hashes_removed);
		int32_t cur_idx = 0;
		for(auto& kvp : hcounts) {
			if(kvp.second <= max_labs_for_kmer) {
				auto& e = table[kvp.first];
				if(kvp.second > 1) {
					e.idx = cur_idx;
					cur_idx += kvp.second;
					e.num_labs = 0;
				}
				else
					e.idx = -1;
			}

			if(verbose)
				++*progress_bar;
		}

		if(verbose) {
			progress_bar->done();
			progress_bar.reset(new progresscpp::ProgressBar(num_labs, 100, "Filling compact index"));
		}

		multipliers.resize(4);
		for(int p = 1; p < 4; ++p)
			multipliers[p] = pow(15.0, p);

#pragma omp parallel for
		for(int l = 0; l < num_labs; ++l) {
			for(auto& kvp : tables[l]) {
				auto& e = table[kvp.first];
				if(e.idx == -1) {
					e.idx = -l - 1;
					e.num_labs = kvp.second;
				}
				else {
					uint32_t nl = e.num_labs++;
					counts[e.idx + nl].lab = l;
					counts[e.idx + nl].weight = (double)kvp.second * 0.0001 * (hcounts[kvp.first] < 4 ? multipliers[4 - hcounts[kvp.first]] : 1.0);
				}
			}

			if(verbose)
				++*progress_bar;
		}

		if(verbose)
			progress_bar->done();
	}

	hash_lookup(istream& file) : hasher(file) {
		serialize_sized(file, ks);
		serialize_sized(file, counts);
		int table_size;
		serialize_value(file, table_size);
		table.reserve(table_size);
		for(int i = 0; i < table_size; ++i) {
			hash_t h;
			serialize_value(file, h);
			hash_entry e;
			serialize_value(file, e);
			table[h] = e;
		}

		multipliers.resize(4);
		for(int p = 1; p < 4; ++p)
			multipliers[p] = pow(15.0, p);
	}

	size_t size() const {
		return table.size();
	}

	void subtract(const vector<string>& seqs) {
		vector<hash_t> to_remove;
		mutex remove_mutex;
#pragma omp parallel for
		for(int t = 0; t < (int)seqs.size(); ++t) {
			hasher.for_all_kmers(seqs[t], ks, [this, &remove_mutex, &to_remove](hash_t h, int p, int size){
				auto it = table.find(h);
				if(it != table.end()) {
					lock_guard<mutex> g(remove_mutex);
					to_remove.push_back(h);
				}
			});
		}

		for(hash_t h : to_remove)
			table.erase(h);
	}

	int num_hashes(int sequence_length) const {
		int hs = 0;
		for(int k : ks)
			hs += max(0, sequence_length - k + 1);

		return hs;
	}

	vector<double> match2(const string& s, int max_num_labs) const {
		vector<double> scores(num_labs, 0.0);
//		unordered_set<hash_t> seen_hashes;
		hasher.for_all_kmers(s, ks, [this, &scores, max_num_labs](hash_t h, int p, int size){
//			if(seen_hashes.count(h) != 0)
//				return;

			auto it = table.find(h);
			if(it == table.end() || (int)it->second.num_labs > max_num_labs)
				return;

//			seen_hashes.insert(h);

			if(it->second.idx >= 0)
				for(int i = 0; i < (int)it->second.num_labs; ++i) {
					auto& c = counts[it->second.idx + i];
					scores[c.lab] += box_cox_transform(1 + c.weight);
				}
			else
				scores[-it->second.idx - 1] += box_cox_transform(1 + it->second.num_labs * 0.0001 * multipliers[3]);
		});

		return scores;
	}

	vector<double> match_leave_one_out(const string& s, uint32_t lab, int max_num_labs) const {
		vector<double> scores(num_labs, 0.0);
		hasher.for_all_kmers(s, ks, [this, &scores, max_num_labs, lab](hash_t h, int p, int size){
			auto it = table.find(h);
			if(it == table.end())
				return;

			int actual_num_labs = it->second.num_labs;
			double weight_adjustment = 1.0;
			double inv_scale = 1.0 / (0.0001 * (it->second.num_labs < 4 ? multipliers[4 - it->second.num_labs] : 1.0));
			if(it->second.idx >= 0)
				for(int i = 0; i < (int)it->second.num_labs; ++i) {
					auto& c = counts[it->second.idx + i];
					if(c.lab == lab) {
						int num_seqs = (int)(c.weight * inv_scale + 0.5);
						if(num_seqs == 1) {
							--actual_num_labs;
							weight_adjustment = inv_scale * 0.0001 * (actual_num_labs < 4 ? multipliers[4 - actual_num_labs] : 1.0);
						}

						break;
					}
				}
			else {
				uint32_t c_lab = -it->second.idx - 1;
				if(c_lab == lab && it->second.num_labs == 1)
					return;
			}

			if(actual_num_labs > max_num_labs)
				return;

			if(it->second.idx >= 0)
				for(int i = 0; i < (int)it->second.num_labs; ++i) {
					auto& c = counts[it->second.idx + i];
					if(c.lab != lab)
						scores.at(c.lab) += c.weight * weight_adjustment;
					else {
						int num_seqs = (int)(c.weight * inv_scale + 0.5);
						double actual_weight = (num_seqs - 1) * 0.0001 * (actual_num_labs < 4 ? multipliers[4 - actual_num_labs] : 1.0);
						scores.at(c.lab) += box_cox_transform(1 + actual_weight);
					}
				}
			else {
				uint32_t c_lab = -it->second.idx - 1;
				int num_seqs = c_lab == lab ? it->second.num_labs - 1 : (int)it->second.num_labs;
				scores[c_lab] += box_cox_transform(1 + num_seqs * 0.0001 * multipliers[3]);
			}
		});

		return scores;
	}

	void save(ostream& file) const {
		hasher.save(file);
		serialize_sized(file, ks);
		serialize_sized(file, counts);
		int table_size = table.size();
		serialize_value(file, table_size);
		for(auto& kvp : table) {
			serialize_value(file, kvp.first);
			serialize_value(file, kvp.second);
		}
	}
private:

};
