#include <iostream>
#include <cinttypes>
#include <vector>
#include <unordered_map>
#include <string>
#include <algorithm>
#include <fstream>
#include <sstream>
#include <iterator>
#include <cmath>
#include <mutex>
#include <numeric>
#include <random>
#include <memory>
#include <unordered_set>
#include <functional>
#include <map>
#include <cctype>

//#include <eigen3/Eigen/Core>

#include "ProgressBar.hpp"
#include "db.h"


using namespace std;
using namespace std::placeholders;


template <typename Out>
void split(const std::string &s, char delim, Out result) {
    std::istringstream iss(s);
    std::string item;
    while (std::getline(iss, item, delim)) {
        *result++ = item;
    }
}


std::vector<std::string> split(const std::string &s, char delim) {
    std::vector<std::string> elems;
    split(s, delim, std::back_inserter(elems));
    return elems;
}


template<class Func>
void for_csv_fields(const string& file_name, Func func, bool skip_header=true) {
	ifstream file(file_name);
	if(!file.is_open())
		throw runtime_error("Cannot open " + file_name);

	cout << "Loading " << file_name << endl;
	string line;
	vector<string> fields;
	if(skip_header)
		getline(file, line);

	while (getline(file, line)) {
		fields.clear();
	    split(line, ',', back_inserter(fields));
	    func(fields);
	}
}


template<class Func>
void for_fasta(const string& file_name, Func func) {
	ifstream file(file_name);
	if(!file.is_open())
		throw runtime_error("Cannot open " + file_name);

	cout << "Loading " << file_name << endl;
	string seq, line;
	while (getline(file, line)) {
		if(line[0] == '>') {
			if(!seq.empty())
				func(seq);

			seq.clear();
		}
		else
			seq += line;
	}

	if(!seq.empty())
		func(seq);
}


void normalize_seq(string& dna_seq) {
	transform(dna_seq.begin(), dna_seq.end(), dna_seq.begin(), [](char c){ return toupper(c); });
	for(char& c : dna_seq)
		switch(c) {
		case 'A':
		case 'C':
		case 'G':
		case 'T':
		case 'N':
			continue;
		default:
			c = 'N';
		}
}


void predict_and_save(const hash_lookup& hl, const vector<string>& test_sequences, const vector<string>& test_ids, int max_labs_per_kmer, const string& file_name, bool normalize,
		const unordered_map<string, uint32_t>& pl_labs) {
	unique_ptr<progresscpp::ProgressBar> progress_bar;
	progress_bar.reset(new progresscpp::ProgressBar(test_sequences.size(), 100, "Computing predictions"));

	vector<vector<double>> test_result(test_sequences.size());
	atomic<int> l1o_count(0);
#pragma omp parallel for
	for(int t = 0; t < (int)test_sequences.size(); ++t) {
		auto s = test_sequences[t];
		auto it = pl_labs.find(test_ids[t]);
		vector<double> scores;
		if(it == pl_labs.end())
			scores = hl.match2(s, max_labs_per_kmer);
		else {
			scores = hl.match_leave_one_out(s, it->second, max_labs_per_kmer);
			++l1o_count;
		}

		if(normalize) {
			double scores_sum = 0;
			for(double s : scores)
				scores_sum += s;

			if(scores_sum == 0.0)
				scores_sum = 1.0;

			for(double& s : scores)
				s /= scores_sum;
		}

		for(double s : scores)
			test_result[t].push_back(s);

		++*progress_bar;
	}

	progress_bar->done();

	if(!pl_labs.empty())
		cout << l1o_count << " samples were seen in the training data and predictions have been adjusted accordingly" << endl;

	ifstream train_lab_file("data/train_labels.csv");
	ofstream test_lab_file(file_name);
	string header;
	getline(train_lab_file, header);
	test_lab_file << header << "\n";
	for(int t = 0; t < (int)test_sequences.size(); ++t) {
		test_lab_file << test_ids[t] << ",";
		for(int l = 0; l < num_labs; ++l) {
			if(test_result[t][l] == 0.0)
				test_lab_file << "0.0";
			else
				test_lab_file << test_result[t][l];

			if(l < num_labs - 1)
				test_lab_file << ",";
		}

		test_lab_file << "\n";
	}
}


void solve(const vector<string>& train_sequences, const vector<uint32_t>& train_labs, const string& output_file) {
	int k0 = 19;
	int max_labs_per_kmer = 18;
	hash_lookup hl(train_sequences, train_labs, {k0, k0 + 2, k0 + 4, k0 + 6, k0 + 8}, max_labs_per_kmer, true);
	cout << hl.size() << " hashes in db" << endl;

	vector<string> test_sequences, test_ids;
	for_csv_fields("data/test_values.csv", [&test_sequences, &test_ids](const vector<string>& fields){
		test_sequences.push_back(fields[1]);
		test_ids.push_back(fields[0]);
	});

	cout << test_sequences.size() << " test samples" << endl;
	predict_and_save(hl, test_sequences, test_ids, max_labs_per_kmer, output_file, false, {});
}


void save_matrix(const vector<vector<double>>& data, const string& file_name) {
	cout << "Saving " << data.size() << " x " << data[0].size() << " matrix into " << file_name << endl;
	ofstream file(file_name);
	if(!file.is_open())
		throw runtime_error("Could not open " + file_name + " for writing");

	for(int t = 0; t < (int)data.size(); ++t) {
		for(int l = 0; l < num_labs; ++l) {
			if(data[t][l] == 0.0)
				file << "0.0";
			else
				file << data[t][l];

			if(l < num_labs - 1)
				file << "\t";
		}

		file << "\n";
	}
}


void add_prediction_on_batch(const hash_lookup& hl, const vector<string>& seqs, vector<double>& scores, int max_labs_per_kmer) {
	vector<vector<double>> all_scores(seqs.size());

#pragma omp parallel for
	for(int t = 0; t < (int)seqs.size(); ++t) {
		auto s = seqs[t];
		all_scores[t] = hl.match2(s, max_labs_per_kmer);
	}

	for(auto& sc : all_scores)
		for(int l = 0; l < num_labs; ++l)
			scores[l] += sc[l];
}


int main(int argc, char **argv) {
	int k0 = 19;
	int max_labs_per_kmer = 18;
	vector<int> ks {k0, k0 + 2, k0 + 4, k0 + 6, k0 + 8};
	int batch_size = 10000;

	if(argc >= 3) {
		string command = argv[1];
		if(command == "build") {
			vector<string> train_sequences;
			vector<string> ids;
			vector<uint32_t> train_labs;
			int i = 2;
			for(; i + 2 < argc; i += 2) {
				for_csv_fields(argv[i], [&train_sequences, &ids, i](const vector<string>& fields){
					ids.push_back(fields[0] + "." + to_string(i / 2));
					train_sequences.push_back(fields[1]);
				});

				for_csv_fields(argv[i + 1], [&train_labs](const vector<string>& fields) {
					for(int j = 0; j < num_labs; ++j)
						if(fields[j + 1] != "0.0") {
							train_labs.push_back(j);
							break;
						}
				}, true);

				if(train_sequences.size() != train_labs.size())
					throw runtime_error("Input features and labels are of different length");
			}

			cout << train_sequences.size() << " train sequences" << endl;
			hash_lookup hl(train_sequences, train_labs, ks, max_labs_per_kmer + 1, true);
			cout << hl.size() << " hashes in db" << endl;

			cout << "Saving DB in " << argv[i] << endl;
			ofstream db_file(argv[i]);
			if(!db_file.is_open())
				throw runtime_error("Could not open " + string(argv[i]));

			hl.save(db_file);
			int num_ids = ids.size();
			serialize_value(db_file, num_ids);
			for(int j = 0; j < num_ids; ++j) {
				serialize_sized(db_file, ids[j]);
				serialize_value(db_file, train_labs[j]);
			}

			cout << "Done." << endl;
			return 0;
		}

		if(command == "predict") {
			cout << "Loading DB from " << argv[2] << endl;
			ifstream db_file(argv[2]);
			if(!db_file.is_open())
				throw runtime_error("Could not open " + string(argv[2]));

			hash_lookup hl(db_file);
			cout << hl.size() << " hashes in db" << endl;

			unordered_map<string, uint32_t> train_ids;
			int a = 3;
			string id_suffix;
			if(argv[a] == string("-")) {
				cout << "Ignoring sample IDs in the input. Make sure that the input does not intersect the training set" << endl;
				++a;
			}
			else {
				id_suffix = string(".") + argv[a];
				++a;
				int num_ids;
				serialize_value(db_file, num_ids);
				train_ids.reserve(num_ids);
				for(int j = 0; j < num_ids; ++j) {
					string id;
					serialize_sized(db_file, id);
					uint32_t lab;
					serialize_value(db_file, lab);
					if(train_ids.count(id) != 0)
						cout << "Duplicate ID in DB: " << id << endl;

					train_ids[id] = lab;
				}
			}

			vector<string> test_sequences, test_ids;
			for_csv_fields(argv[a], [&test_sequences, &test_ids, &id_suffix](const vector<string>& fields){
				test_sequences.push_back(fields[1]);
				test_ids.push_back(fields[0] + id_suffix);
			});

			cout << test_sequences.size() << " test samples" << endl;
			predict_and_save(hl, test_sequences, test_ids, max_labs_per_kmer, argv[a + 1], false, train_ids);
			cout << "Done." << endl;
			return 0;
		}

		if(command == "predict_fasta") {
			cout << "Loading DB from " << argv[2] << endl;
			ifstream db_file(argv[2]);
			if(!db_file.is_open())
				throw runtime_error("Could not open " + string(argv[2]));

			hash_lookup hl(db_file);
			cout << hl.size() << " hashes in db" << endl;

			vector<double> scores(num_labs, 0.0);
			vector<string> seqs_batch;
			for_fasta(argv[3], [&hl, &scores, &seqs_batch, max_labs_per_kmer, batch_size](string& seq){
				normalize_seq(seq);
				seqs_batch.push_back(seq);
				if((int)seqs_batch.size() >= batch_size) {
					add_prediction_on_batch(hl, seqs_batch, scores, max_labs_per_kmer);
					seqs_batch.clear();
				}
			});

			add_prediction_on_batch(hl, seqs_batch, scores, max_labs_per_kmer);
			save_matrix({scores}, argv[4]);

			cout << "Done." << endl;
			return 0;
		}

		if(command == "subtract") {
			int subtract_batch_size = 16;
			cout << "Loading DB from " << argv[2] << endl;
			ifstream db_file(argv[2]);
			if(!db_file.is_open())
				throw runtime_error("Could not open " + string(argv[2]));

			hash_lookup hl(db_file);
			db_file.close();
			size_t prev_size = hl.size();
			cout << prev_size << " hashes in db" << endl;

			vector<string> seqs_batch;
			for_fasta(argv[3], [&hl, &seqs_batch, subtract_batch_size](string& seq){
				normalize_seq(seq);
				seqs_batch.push_back(seq);
				if((int)seqs_batch.size() >= subtract_batch_size) {
					hl.subtract(seqs_batch);
					seqs_batch.clear();
				}
			});

			hl.subtract(seqs_batch);

			cout << prev_size - hl.size() << " hashes removed. " << hl.size() << " remain" << endl;

			cout << "Saving DB in " << argv[4] << endl;
			ofstream db_file_out(argv[4]);
			if(!db_file_out.is_open())
				throw runtime_error("Could not open " + string(argv[4]));

			hl.save(db_file_out);
			int num_ids = 0;
			serialize_value(db_file_out, num_ids);

			cout << "Done." << endl;
			return 0;
		}
	}

	cout << "Usage:\nhash_match build <in_features1> <in_labels1> [<in_features2> <in_labels2>] ... <out_db_file>\n" <<
			"hash_match predict <in_db_file> (-|<position of this dataset in the inputs of build commands>) <in_features> <out_predictions>\n" <<
			"hash_match predict_fasta <in_db_file> <in_fasta> <out_prediction>\n" <<
			"hash_match subtract <in_db_file> <in_fasta> <out_db_file>\n" <<
			"The arguments must be provided in the indicated order" << endl;

	return 1;
}
