import os

import numpy as np
from tqdm import trange
import torch
import torch.nn as nn
from torch.cuda.amp import GradScaler, autocast

from archs import Pool2BN, SimpleCNN, MLP, JoinNet, ResNet1d, NTXentLoss

torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False

num_labs = 1314

repr_model_file = f"models/ssrepr.model"
repr_model_file_r3 = f"models/ssrepr_r3.model"


def check_cuda(gpu):
    assert torch.cuda.is_available(), "No GPU available as reported by pytorch. You need a GPU to use this model"
    print(f"{torch.cuda.device_count()} GPUs found")
    assert 0 <= gpu < torch.cuda.device_count(), f"Selected GPU #{gpu}, but it's not available"


def seq_minibatch(seqs, train_sequence_length, cuda_dev, drop_prob=None, random_rc=False, mut_prob=None):
    mb = torch.zeros((len(seqs), 4, train_sequence_length), dtype=torch.float32, device=cuda_dev)
    for i, s in enumerate(seqs):
        s = s.to(cuda_dev)

        if mut_prob is not None:
            s = s.clone()
            to_mutate = int(len(s) * mut_prob)
            s[np.random.choice(len(s), to_mutate)] = torch.LongTensor(np.random.randint(4, size=to_mutate))

        # circular_shift = np.random.randint(len(s))
        # s = torch.cat((s[circular_shift:], s[:circular_shift]))
        if len(s) <= train_sequence_length:
            content = torch.nn.functional.one_hot(s, 5)[:, :4].transpose(0, 1)
            # for offset in range(0, train_sequence_length, len(s)):
            #    l = min(len(s), train_sequence_length - offset)
            #    mb[i, :, offset:offset + l] = content[:, :l]
            mb[i, :, :len(s)] = content
        else:
            offset = np.random.randint(0, len(s) - train_sequence_length + 1)
            mb[i, ...] = torch.nn.functional.one_hot(s[offset:offset + train_sequence_length], 5)[:, :4].transpose(0, 1)

        if drop_prob is not None:
            mb[i, :, np.random.choice(train_sequence_length, int(train_sequence_length * drop_prob))] = 0

        if random_rc and np.random.randint(2) == 0:
            mb[i, ...] = torch.flip(mb[i, ...], [0, 1])

    return mb


def seq_batch_fullsize(seq, train_sequence_length, cuda_dev):
    content = torch.nn.functional.one_hot(seq, 5)[:, :4].transpose(0, 1)
    if len(seq) <= train_sequence_length:
        mb = torch.zeros((1, 4, train_sequence_length), dtype=torch.float32, device=cuda_dev)
        mb[0, :, :len(seq)] = content
    else:
        stride = train_sequence_length // 2
        splits = (len(seq) - 1) // stride + 1
        mb = torch.zeros((splits, 4, train_sequence_length), dtype=torch.float32, device=cuda_dev)
        for i in range(splits):
            offset = i * stride
            if offset + train_sequence_length >= len(seq):
                # length_until_end = len(seq) - offset
                # mb[i, :, :length_until_end] = content[:, offset:]
                # mb[i, :, length_until_end:] = content[:, :(train_sequence_length - length_until_end)]
                mb[i, ...] = content[:, -train_sequence_length:]
            else:
                mb[i, ...] = content[:, offset:offset + train_sequence_length]

    return torch.cat((mb, torch.flip(mb, [1, 2])), dim=0)


def labels_idx(train_labels):
    label_to_idx = [[] for _ in range(num_labs)]
    for i, l in enumerate(train_labels):
        label_to_idx[l].append(i)

    return label_to_idx


def correct_topk(output, target, k):
    _, pred = output.topk(k, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))
    correct_k = correct[:k].view(-1).float().sum(0)
    return correct_k.item()


def predict_aug(model, x, batch_size, train_sequence_length):
    preds = []
    with torch.no_grad():
        model.eval()
        for batch_idx in range(0, len(x[0]), batch_size):
            batch_seq = seq_minibatch(x[0][batch_idx:batch_idx + batch_size], train_sequence_length, model.device)
            batch_bin = torch.from_numpy(x[1][batch_idx:batch_idx + batch_size, ...]).to(model.device)
            preds.append(model(batch_seq, batch_bin))

    return torch.cat(preds, axis=0)


def ew_idx_75(idx, to_take, n_samples):
    selected_labels = np.random.choice(num_labs, to_take // 4 * 3, replace=True)
    selected_idx = [np.random.choice(idx[l]) for l in selected_labels if idx[l]]
    return selected_idx + list(np.random.choice(n_samples, to_take // 4))


def ew_idx_100(idx, to_take, _):
    selected_labels = np.random.randint(num_labs, size=to_take)
    selected_idx = [np.random.choice(idx[l]) for l in selected_labels if idx[l]]
    return selected_idx


def make_trainer_aug_ew(sample_indexer, **kwargs):
    def trainer_aug(model, train_x, train_y, batch_idx, criterion):
        f_seq, f_bin, f_idx = train_x
        selected_idx = sample_indexer(f_idx, len(batch_idx), len(f_seq))
        train_batch_seq = seq_minibatch([f_seq[i] for i in selected_idx], cuda_dev=model.device, **kwargs)
        train_batch_bin = torch.from_numpy(f_bin[selected_idx, ...]).to(model.device)
        preds = model(train_batch_seq, train_batch_bin)
        return criterion(preds, torch.from_numpy(train_y[selected_idx, ...]).to(model.device))

    return trainer_aug


def make_trainer_aug(**kwargs):
    def trainer_aug(model, train_x, train_y, batch_idx, criterion):
        train_batch_seq = seq_minibatch([train_x[0][i] for i in batch_idx], cuda_dev=model.device, **kwargs)
        train_batch_bin = torch.from_numpy(train_x[1][batch_idx, ...]).to(model.device)
        preds = model(train_batch_seq, train_batch_bin)
        return criterion(preds, torch.from_numpy(train_y[batch_idx, ...]).to(model.device))

    return trainer_aug


def top10_score_aug(model, x, y, batch_size, train_sequence_length):
    return correct_topk(predict_aug(model, x, batch_size, train_sequence_length), torch.from_numpy(y).to(model.device),
                        10) / len(y)


def predict_full_aug(model, x, train_sequence_length):
    preds = []
    with torch.no_grad():
        model.eval()
        for batch_idx in trange(len(x[0])):
            batch_seq = seq_batch_fullsize(x[0][batch_idx], train_sequence_length, model.device)
            batch_bin = \
                torch.from_numpy(x[1][batch_idx:batch_idx + 1, ...]).expand(batch_seq.shape[0], -1).to(model.device)

            preds.append(model(batch_seq, batch_bin).mean(dim=0, keepdim=True))

    return torch.cat(preds, axis=0)


def make_train_batch(train_seqs, lab_idx, to_take, cuda_dev, **kwargs):
    labs = [l for l in np.random.randint(num_labs, size=to_take) if lab_idx[l]]
    seqs_targets = [(np.random.choice(lab_idx[l]), np.random.choice(lab_idx[l])) for l in labs]
    seqs = []
    for i, (s1, s2) in enumerate(seqs_targets):
        if s1 == s2:
            break_point = np.random.randint(5, len(train_seqs[s1]) - 5)
            seqs.append(train_seqs[s1][:break_point])
            seqs.append(train_seqs[s1][break_point:])
        else:
            seqs.append(train_seqs[s1])
            seqs.append(train_seqs[s2])

    labs2 = []
    for l in labs:
        labs2.append(l)
        labs2.append(l)

    batch = seq_minibatch(seqs, cuda_dev=cuda_dev, **kwargs)
    targets = torch.from_numpy(np.array(labs2, dtype=np.int64)).to(cuda_dev)
    return batch, targets


def make_test_batch(test_seqs, to_take, cuda_dev, **kwargs):
    seqs_chosen = np.random.choice(len(test_seqs), to_take, replace=True)
    seqs = []
    labs2 = []
    for s in seqs_chosen:
        break_point = np.random.randint(5, len(test_seqs[s]) - 5)
        seqs.append(test_seqs[s][:break_point])
        seqs.append(test_seqs[s][break_point:])

        labs2.append(-s - 1)
        labs2.append(-s - 1)

    batch = seq_minibatch(seqs, cuda_dev=cuda_dev, **kwargs)
    targets = torch.from_numpy(np.array(labs2, dtype=np.int64)).to(cuda_dev)
    return batch, targets


def make_trainer_ss(**kwargs):
    def trainer(model, train_x, train_y, batch_idx, criterion):
        f_seq, f_idx = train_x
        to_take = len(batch_idx)
        batch_inputs, batch_labels = make_train_batch(f_seq, f_idx, to_take, cuda_dev=model.device, **kwargs)
        reprs = model(batch_inputs)
        return criterion(reprs, batch_labels)

    return trainer


def fit(model, opt, criterion, scheduler, scaler, train_x, train_y, epochs, batch_size,
        model_file, trainer, n_train_samples=None,
        autocast_enabled=True):
    epoch_loss_train = []

    if n_train_samples is None:
        n_train_samples = train_x.shape[0]

    t = trange(epochs)
    for _ in t:
        rnd_idx = list(range(n_train_samples))
        np.random.shuffle(rnd_idx)
        losses = []
        for batch_idx in range(0, n_train_samples, batch_size):
            model.train()
            opt.zero_grad()
            with autocast(enabled=autocast_enabled):
                loss = trainer(model, train_x, train_y, rnd_idx[batch_idx:(batch_idx + batch_size)], criterion)

            scaler.scale(loss).backward()
            scaler.step(opt)
            scaler.update()
            losses.append(loss.item())

        epoch_loss_train.append(np.mean(losses))
        scheduler.step()

        t.set_description(f"train: {epoch_loss_train[-1]:.3f}")

    torch.save(model.state_dict(), model_file)
    model.eval()
    return model, epoch_loss_train


def make_model1(num_aug_features):
    hidden_dim = 2000
    width = 2000
    seq_drop_prob = 0.05
    train_sequence_length = 8000
    weight_decay = 0.0001

    model_bin = MLP([num_aug_features, hidden_dim, hidden_dim, num_labs], dropout=0.5)
    bin_headless = model_bin.mlp[:7]

    model_seq = SimpleCNN(18, hidden_dim, additional_layer=False)

    new_head = torch.nn.Sequential(
        torch.nn.Dropout(0.5),
        MLP([width * 2 + hidden_dim, num_labs])
    )

    joined_model = JoinNet(model_seq, bin_headless, new_head)

    return joined_model, train_sequence_length, seq_drop_prob, weight_decay


def make_model2(num_aug_features):
    hidden_dim = 2048
    width = 2048
    seq_drop_prob = 0.05
    train_sequence_length = 16384
    weight_decay = 0.0001

    model_bin = MLP([num_aug_features, hidden_dim, hidden_dim, num_labs], dropout=0.5)
    bin_headless = model_bin.mlp[:7]

    model_seq = SimpleCNN(18, hidden_dim, additional_layer=False)

    new_head = torch.nn.Sequential(
        torch.nn.Dropout(0.5),
        MLP([width * 2 + hidden_dim, num_labs])
    )

    joined_model = JoinNet(model_seq, bin_headless, new_head)

    return joined_model, train_sequence_length, seq_drop_prob, weight_decay


def make_model2_old(num_aug_features):
    hidden_dim = 2000
    width = 1000
    seq_drop_prob = 0.05
    train_sequence_length = 8050
    weight_decay = 0.0001

    model_bin = MLP([num_aug_features, hidden_dim, hidden_dim, num_labs], dropout=0.5)
    bin_headless = model_bin.mlp[:7]

    model_seq = SimpleCNN(18, width, additional_layer=True)

    new_head = torch.nn.Sequential(
        torch.nn.Dropout(0.5),
        MLP([width * 2 + hidden_dim, num_labs])
    )

    joined_model = JoinNet(model_seq, bin_headless, new_head)

    return joined_model, train_sequence_length, seq_drop_prob, weight_decay


def make_model3(num_aug_features):
    hidden_dim = 2048
    width = 1024
    seq_drop_prob = 0.05
    train_sequence_length = 8196
    weight_decay = 0.0001

    model_bin = MLP([num_aug_features, hidden_dim, hidden_dim, num_labs], dropout=0.5)
    bin_headless = model_bin.mlp[:7]

    model_seq = SimpleCNN(18, width, additional_layer=True)

    new_head = torch.nn.Sequential(
        torch.nn.Dropout(0.5),
        MLP([width * 2 + hidden_dim, num_labs])
    )

    joined_model = JoinNet(model_seq, bin_headless, new_head)

    return joined_model, train_sequence_length, seq_drop_prob, weight_decay


def make_model4(num_aug_features):
    seq_drop_prob = 0.05
    hidden_dim = 2000
    width = 768
    train_sequence_length = 8100
    weight_decay = 0.0001

    model_bin = MLP([num_aug_features, hidden_dim, hidden_dim, num_labs], dropout=0.5)
    bin_headless = model_bin.mlp[:7]

    model_seq = nn.Sequential(
        nn.Conv1d(4, width, 7, padding=3),
        nn.ReLU(),
        nn.BatchNorm1d(width),
        ResNet1d(width, [(3, width // 2, 1)] * 1, dropout=None, dilation=7),
        nn.ReLU(),
        Pool2BN(width),
    )

    new_head = torch.nn.Sequential(
        torch.nn.Dropout(0.5),
        MLP([width * 2 + hidden_dim, num_labs])
    )

    joined_model = JoinNet(model_seq, bin_headless, new_head)
    return joined_model, train_sequence_length, seq_drop_prob, weight_decay


def make_model5(num_aug_features):
    seq_drop_prob = 0.05
    hidden_dim = 2048
    width = 768
    train_sequence_length = 8128
    weight_decay = 0.0001

    model_bin = MLP([num_aug_features, hidden_dim, hidden_dim, num_labs], dropout=0.5)
    bin_headless = model_bin.mlp[:7]

    model_seq = nn.Sequential(
        nn.Conv1d(4, width, 7, padding=3),
        nn.ReLU(),
        nn.BatchNorm1d(width),
        ResNet1d(width, [(3, width // 2, 1)] * 1, dropout=0.1, dilation=7),
        nn.ReLU(),
        Pool2BN(width),
    )

    new_head = torch.nn.Sequential(
        torch.nn.Dropout(0.5),
        MLP([width * 2 + hidden_dim, num_labs])
    )

    joined_model = JoinNet(model_seq, bin_headless, new_head)
    return joined_model, train_sequence_length, seq_drop_prob, weight_decay


def make_ss_embeddings_model():
    hidden_dim = 2000
    width = 2000
    seq_drop_prob = 0.05
    train_sequence_length = 8000
    weight_decay = 0.0001

    repr_model = torch.nn.Sequential(
        SimpleCNN(18, width, additional_layer=False),
        MLP([width * 2, hidden_dim, hidden_dim])
    )

    return repr_model, train_sequence_length, seq_drop_prob, weight_decay


def make_model_ss(num_aug_features, rare_multiply):
    hidden_dim = 2000
    width = 2000
    repr_model, train_sequence_length, seq_drop_prob, weight_decay = make_ss_embeddings_model()
    repr_model.load_state_dict(torch.load(repr_model_file_r3 if rare_multiply else repr_model_file))

    model_bin = MLP([num_aug_features, hidden_dim, hidden_dim, num_labs], dropout=0.5)
    bin_headless = model_bin.mlp[:7]

    new_head = torch.nn.Sequential(
        torch.nn.Dropout(0.5),
        MLP([width + hidden_dim, hidden_dim, num_labs])
    )

    joined_model = JoinNet(repr_model, bin_headless, new_head)
    return joined_model, train_sequence_length, seq_drop_prob, weight_decay


models_list = [make_model1, make_model2, make_model3, make_model4, make_model5, lambda n: make_model_ss(n, False),
               lambda n: make_model_ss(n, True)]


def train_aug_nn_model(model_number, train_seqs, train_bin_features, train_labels, cuda_dev):
    cuda_dev = torch.device(cuda_dev)

    seed = 30 + model_number
    model_file = f"models/nn3_seed{seed}.model"

    np.random.seed(seed)
    torch.manual_seed(seed)

    train_lidx = labels_idx(train_labels)

    model_maker = models_list[model_number]
    joined_model, train_sequence_length, seq_drop_prob, weight_decay = model_maker(train_bin_features.shape[1])
    joined_model = joined_model.to(cuda_dev)
    joined_model.device = cuda_dev

    opt = torch.optim.Adam(joined_model.parameters(), lr=0.001, weight_decay=weight_decay)
    criterion = torch.nn.CrossEntropyLoss(reduction="mean")

    scheduler = torch.optim.lr_scheduler.ExponentialLR(opt, 0.97)

    scaler = GradScaler()
    indexer = ew_idx_75

    batch_size = 128 if train_sequence_length < 8500 else 64

    joined_model_trained, epoch_loss_train = \
        fit(joined_model, opt, criterion, scheduler,
            scaler,
            (train_seqs, train_bin_features, train_lidx),
            train_labels,
            epochs=100,
            batch_size=batch_size,
            model_file=model_file,
            trainer=make_trainer_aug_ew(sample_indexer=indexer,
                                        train_sequence_length=train_sequence_length,
                                        drop_prob=seq_drop_prob,
                                        random_rc=True, mut_prob=None),
            n_train_samples=len(train_labels))

    print(f"Min training loss={min(epoch_loss_train)}")


def train_self_supervised_embeddings_model(train_seqs, train_labels, cuda_dev, rare_multiply, seed):
    cuda_dev = torch.device(cuda_dev)

    np.random.seed(seed)
    torch.manual_seed(seed)

    train_lidx = labels_idx(train_labels)

    if rare_multiply:
        train_lidx.sort(key=lambda l: len(l))
        train_lidx += train_lidx[:num_labs // 2] * 2

    repr_model, train_sequence_length, seq_drop_prob, weight_decay = make_ss_embeddings_model()
    repr_model = repr_model.to(cuda_dev)

    repr_model.device = cuda_dev

    opt = torch.optim.Adam(repr_model.parameters(), lr=0.001, weight_decay=weight_decay)
    criterion = NTXentLoss(0.5, True)
    scheduler = torch.optim.lr_scheduler.ExponentialLR(opt, 0.97)
    scaler = GradScaler()
    repr_model_trained, epoch_loss_train = fit(repr_model, opt, criterion, scheduler,
                                               scaler,
                                               (train_seqs, train_lidx), train_labels,
                                               epochs=100,
                                               batch_size=64,
                                               model_file=repr_model_file_r3 if rare_multiply else repr_model_file,
                                               trainer=make_trainer_ss(
                                                   train_sequence_length=train_sequence_length,
                                                   drop_prob=seq_drop_prob, random_rc=True,
                                                   mut_prob=None),
                                               n_train_samples=len(train_labels))

    print(f"Min training loss={min(epoch_loss_train)}")


def train_aug_ss_nn_model(model_number, train_seqs, train_bin_features, train_labels, cuda_dev):
    rare_multiply = model_number == 6
    if os.path.exists(repr_model_file_r3 if rare_multiply else repr_model_file):
        print("Embedding model found, will reuse")
    else:
        print("Training self-supervised embedding model" + ": r3" * rare_multiply)
        train_self_supervised_embeddings_model(train_seqs, train_labels, cuda_dev, rare_multiply, 50 + model_number)

    print("Training augmented model using self-supervised embedding model as initial point")
    train_aug_nn_model(model_number, train_seqs, train_bin_features, train_labels, cuda_dev)


def predict_nn_model(model_number, test_seqs, test_bin_features, cuda_dev, base_seed=30, model_file=None):
    cuda_dev = torch.device(cuda_dev)
    seed = base_seed + model_number
    model_maker = models_list[model_number]
    joined_model, train_sequence_length, seq_drop_prob, weight_decay = model_maker(test_bin_features.shape[1])
    joined_model = joined_model.to(cuda_dev)
    joined_model.device = cuda_dev

    if model_file is None:
        model_file = f"models/nn3_seed{seed}.model"

    joined_model.load_state_dict(torch.load(model_file))
    joined_model.eval()

    test_preds = predict_full_aug(joined_model, (test_seqs, test_bin_features),
                                  train_sequence_length).cpu().numpy()

    return test_preds
