import argparse
import subprocess
import os

import preprocess
import nn_model
import ensemble


def run(command):
    try:
        subprocess.check_call(command, shell=True, executable="/bin/bash")
    except Exception:
        assert False, f"Failed command: {command}"


def output_exists(file):
    if isinstance(file, str):
        return os.path.exists(file)

    return all(os.path.exists(f) for f in file)


def make_file(command, output_file, name):
    if output_exists(output_file):
        print(f"{name}: skipping because its output '{output_file}' exists\n", flush=True)
        return

    if isinstance(command, str):
        print(f"{name}\nOutput file: '{output_file}'\nRunning command: {command}", flush=True)
        run(command)
    else:
        print(f"{name}\nOutput file: '{output_file}'", flush=True)
        command()

    assert output_exists(output_file), f"Failed to produce output '{output_file}'"
    print(f"Done: {name}\n", flush=True)


def train_models(only_model, gpu):
    assert os.path.exists("models") and os.path.isdir("models"), "Couldn't find folder 'models/'. I won't create it " \
                                                                 "because if this is run through docker, the folder " \
                                                                 "will be inaccessible. Please create it if running "\
                                                                 "natively or mount it if running through docker"

    assert os.path.exists("data/train_values.csv") and os.path.exists("data/train_labels.csv"), \
           "Couldn't find the training dataset. Please copy or mount it in the 'data/' folder"

    print("""
████████╗██████╗  █████╗ ██╗███╗   ██╗██╗███╗   ██╗ ██████╗ 
╚══██╔══╝██╔══██╗██╔══██╗██║████╗  ██║██║████╗  ██║██╔════╝ 
   ██║   ██████╔╝███████║██║██╔██╗ ██║██║██╔██╗ ██║██║  ███╗
   ██║   ██╔══██╗██╔══██║██║██║╚██╗██║██║██║╚██╗██║██║   ██║
   ██║   ██║  ██║██║  ██║██║██║ ╚████║██║██║ ╚████║╚██████╔╝
   ╚═╝   ╚═╝  ╚═╝╚═╝  ╚═╝╚═╝╚═╝  ╚═══╝╚═╝╚═╝  ╚═══╝ ╚═════╝ 
""", flush=True)

    if only_model is None:
        make_file("./hash_match build data/train_values.csv data/train_labels.csv models/kmer_db.bin",
                  "models/kmer_db.bin", "Build k-mer DB")

        print(f"Training {len(nn_model.models_list)} neural network models. Approximate time to finish: 65 GPU hours.",
              flush=True)

    train_seqs, train_bin_features = preprocess.load_features("data/train_values.csv")
    train_labels = preprocess.load_labels("data/train_labels.csv")

    cuda_dev = f"cuda:{gpu}"
    print(f"Using CUDA device {cuda_dev}", flush=True)

    for model_number in range(len(nn_model.models_list)):
        if only_model is not None and only_model != model_number:
            print(f"Train model #{model_number}: skipping because it is not selected for building", flush=True)
            continue

        seed = 30 + model_number

        def maker():
            if model_number < 5:
                nn_model.train_aug_nn_model(model_number, train_seqs, train_bin_features, train_labels, cuda_dev)
            else:
                nn_model.train_aug_ss_nn_model(model_number, train_seqs, train_bin_features, train_labels,
                                               cuda_dev)

        make_file(maker,
                  f"models/nn3_seed{seed}.model",
                  f"Train model #{model_number}")

    if only_model is None:
        print("Training of all models finished. Now you can predict", flush=True)


def predict(gpu):
    assert os.path.exists("outputs") and os.path.isdir("outputs"), "Couldn't find folder 'outputs/'. I won't create " \
                                                                   "it " \
                                                                   "because if this is run through docker, the " \
                                                                   "folder " \
                                                                   "will be inaccessible. Please create it if " \
                                                                   "running " \
                                                                   "natively or mount it if running through docker"

    assert os.path.exists("models/kmer_db.bin"), "Could not find kmer DB"

    for model_number in range(6):
        seed = 30 + model_number
        assert os.path.exists(f"models/nn3_seed{seed}.model"), f"Could not find trained model #{model_number}"

    assert os.path.exists("data/test_values.csv"), \
           "Couldn't find the test dataset. Please copy or mount it in the 'data/' folder"

    print("""
██████╗ ██████╗ ███████╗██████╗ ██╗ ██████╗████████╗██╗███╗   ██╗ ██████╗ 
██╔══██╗██╔══██╗██╔════╝██╔══██╗██║██╔════╝╚══██╔══╝██║████╗  ██║██╔════╝ 
██████╔╝██████╔╝█████╗  ██║  ██║██║██║        ██║   ██║██╔██╗ ██║██║  ███╗
██╔═══╝ ██╔══██╗██╔══╝  ██║  ██║██║██║        ██║   ██║██║╚██╗██║██║   ██║
██║     ██║  ██║███████╗██████╔╝██║╚██████╗   ██║   ██║██║ ╚████║╚██████╔╝
╚═╝     ╚═╝  ╚═╝╚══════╝╚═════╝ ╚═╝ ╚═════╝   ╚═╝   ╚═╝╚═╝  ╚═══╝ ╚═════╝ 

This will take about 10 minutes.    
""", flush=True)

    kmer_preds_file = "outputs/kmer_test_preds.csv"
    make_file(f"./hash_match predict models/kmer_db.bin - data/test_values.csv {kmer_preds_file}",
              kmer_preds_file, "Make predictions for test set with k-mers")

    test_seqs, test_bin_features = preprocess.load_features("data/test_values.csv")
    nn_preds = []
    cuda_dev = f"cuda:{gpu}"
    print(f"Using CUDA device {cuda_dev}", flush=True)
    for model_number in range(len(nn_model.models_list)):
        print(f"Predicting with model #{model_number}", flush=True)
        nn_preds.append(nn_model.predict_nn_model(model_number, test_seqs, test_bin_features, cuda_dev))

    print(f"Ensemble predictions from {len(nn_preds)} neural networks and k-mer DB", flush=True)
    ensemble.ensemble_predictions(kmer_preds_file, nn_preds)
    print("Submission file is ready", flush=True)


def main():
    parser = argparse.ArgumentParser(description="Build and run genetic attribution detection models for "
                                                 "Genetic Engineering Attribution Challenge on DrivenData")

    parser.add_argument('command', type=str, choices=["train", "predict"], help='what to do')
    parser.add_argument('--only-model', metavar="N", type=int, required=False,
                        help="train only this model [0-6]")

    parser.add_argument('--gpu', metavar="N", type=int, required=False, default=0, help="0-based index of GPU to use")
    args = parser.parse_args()

    nn_model.check_cuda(args.gpu)

    if args.command == "train":
        train_models(args.only_model, args.gpu)
    elif args.command == "predict":
        predict(args.gpu)


if __name__ == "__main__":
    main()
