import pandas as pd
import numpy as np
import os
from pathlib import Path
from .converters import seq_to_ints, add_padding


class DataProvider:
    def __init__(self, dataset_path, mode, seq_size, left_padding_size=10):
        self.dataset_path = dataset_path
        self.prepoc_path = Path(os.path.dirname(self.dataset_path)) / 'preprocessed'
        self.mode = mode
        self.seq_size = seq_size
        self.left_padding_size = left_padding_size
        print('Get values...')
        self.values = self._load_values()
        self.index = self.values.index
        print('Get load_seqs_as_intarray...')
        self.seqs = self._load_seqs_as_intarray(values=self.values, reverse=False)
        print('Get load_seqs_as_intarray...')
        self.seqs_reverse = self._load_seqs_as_intarray(values=self.values, reverse=True)
        print('Get load_initial_cat_features...')
        self.init_cat = self._load_initial_cat_features()
        print('Get load_blast_sort_features...')
        self.blast = self._load_blast_sort_features()
        print('Get load_alen_pca...')
        self.pca_alenlog = self._load_alen_pca(components_count=1024)
        print('Get load_bitscore_pca...')
        self.pca_bitscore = self._load_bitscore_pca(components_count=1024)
        print('Get load_mismatch_pca...')
        self.pca_mismatch = self._load_mismatch_pca(components_count=1024)
        print('Get load_gap_pca...')
        self.pca_gap = self._load_gap_pca(components_count=1024)

    def load_classes_names(self):
        train_labels = pd.read_csv(self.dataset_path / 'train_labels.csv', index_col='sequence_id')
        return train_labels.columns

    def load_labels(self, smooth):
        train_labels = pd.read_csv(self.dataset_path / 'train_labels.csv', index_col='sequence_id')
        train_labels = np.array(train_labels.values).astype('float32')
        train_labels = train_labels * smooth + (1 - smooth) / 1314
        return train_labels

    def _load_values(self):
        return pd.read_csv(self.dataset_path / f'{self.mode}_values.csv', index_col='sequence_id')

    def _load_seqs_as_intarray(self, values, reverse):
        return np.stack(values.sequence
                        .apply(lambda x: seq_to_ints(x, seq_size=self.seq_size, reverse=reverse))
                        .apply(lambda x: add_padding(x,
                                                     seq_size=self.seq_size,
                                                     left_padding_size=self.left_padding_size)).values)

    def _load_initial_cat_features(self):
        return np.stack(self.values.iloc[:, 1:].values)

    def _load_blast_sort_features(self):
        blast_sort = pd.read_csv(self.prepoc_path / f'blast_sort_{self.mode}.csv',
                                 index_col='sequence_id')
        blast_sort_all = pd.DataFrame(data=np.zeros((len(self.index), 1314)),
                                      columns=["cl" + str(i) for i in range(1314)],
                                      index=self.index)
        blast_sort_all.loc[blast_sort.index] = blast_sort
        return blast_sort_all.loc[self.index].values

    def _load_pca(self, filename, components_count):
        pca = pd.read_csv(self.prepoc_path / filename, index_col='sequence_id')
        data = pca.loc[self.index].values
        if components_count > 0:
            data = data[:, :components_count]
        return data

    def _load_alen_pca(self, components_count=-1):
        return self._load_pca(f'pca_alenlog_{self.mode}.csv', components_count)

    def _load_gap_pca(self, components_count=-1):
        return self._load_pca(f'pca_gap_{self.mode}.csv', components_count)

    def _load_bitscore_pca(self, components_count=-1):
        return self._load_pca(f'pca_bitscorelog_{self.mode}.csv', components_count)

    def _load_mismatch_pca(self, components_count=-1):
        return self._load_pca(f'pca_mismatches_{self.mode}.csv', components_count)

    def get_normal_seqs_input(self):
        return self.seqs, self.init_cat, self.blast, self.pca_alenlog, self.pca_gap, self.pca_bitscore, self.pca_mismatch

    def get_reversed_seqs_input(self):
        return self.seqs_reverse, self.init_cat, self.blast, self.pca_alenlog, self.pca_gap, self.pca_bitscore, self.pca_mismatch

    def get_concatenated_seqs_input(self):
        all_seqs = np.concatenate([self.seqs, self.seqs_reverse], axis=0)
        all_init_cat = np.concatenate([self.init_cat, self.init_cat], axis=0)
        all_blast = np.concatenate([self.blast, self.blast], axis=0)
        all_pca_alenlog = np.concatenate([self.pca_alenlog, self.pca_alenlog], axis=0)
        all_pca_gap = np.concatenate([self.pca_gap, self.pca_gap], axis=0)
        all_pca_bitscore = np.concatenate([self.pca_bitscore, self.pca_bitscore], axis=0)
        all_pca_mismatch = np.concatenate([self.pca_mismatch, self.pca_mismatch], axis=0)
        return all_seqs, all_init_cat, all_blast, all_pca_alenlog, all_pca_gap, all_pca_bitscore, all_pca_mismatch
