from tensorflow.keras.models import Model
import importlib
from data_providers.data_provider import DataProvider
import tensorflow as tf
import functools
import numpy as np
import os
from .architectures import create_arch_v1, create_arch_v2, create_arch_v3, create_arch_v4, create_arch_v5


def run(args, config_name):
    if args.float16:
        from tensorflow.keras.mixed_precision import experimental as mixed_precision
        policy = mixed_precision.Policy('mixed_float16')
        mixed_precision.set_policy(policy)
    else:
        print("If GPU supports float16 set --float16 True and it will be faster")

    model_config = importlib.import_module(f'configs.{config_name}').config

    # Create DNN architecture
    arch_version = model_config['arch_version']
    if arch_version == 1:
        create_arch = create_arch_v1
    elif arch_version == 2:
        create_arch = create_arch_v2
    elif arch_version == 3:
        create_arch = create_arch_v3
    elif arch_version == 4:
        create_arch = create_arch_v4
    elif arch_version == 5:
        create_arch = create_arch_v5
    else:
        raise AttributeError('arch_version has to be 1 or 2 or 3')

    seq_size = model_config['seq_size']
    left_padding_size = 10
    model = create_arch(seq_size=seq_size,
                        embs_channels_size=model_config['embs_channels_size'],
                        seq_kernels=model_config['seq_kernels'],
                        left_padding_size=left_padding_size)
    model.summary()

    # Load data
    data_provider = DataProvider(mode='train',
                                 dataset_path=args.dataset_path,
                                 seq_size=seq_size,
                                 left_padding_size=left_padding_size)
    print('Concat...')
    nn_input = data_provider.get_concatenated_seqs_input()
    y = data_provider.load_labels(model_config['smooth'])
    all_y = np.concatenate([y, y], axis=0)

    # Train model
    optim = tf.keras.optimizers.Adam(0.001)
    top10_acc = functools.partial(tf.keras.metrics.top_k_categorical_accuracy, k=10)
    top10_acc.__name__ = 'top10_acc'
    model.compile(optim, 'binary_crossentropy', metrics=['acc', top10_acc])

    def scheduler(epoch, lr):
        if epoch < 15:
            new_lr = lr
        else:
            new_lr = lr * 0.5
        print(new_lr)
        return new_lr

    callback = tf.keras.callbacks.LearningRateScheduler(scheduler)


    saver_callback = tf.keras.callbacks.ModelCheckpoint(str(args.models_path / (config_name + '.h5')))

    model.fit(
        nn_input,
        all_y,
        batch_size=model_config.get('batch_size', 32),
        epochs=21,
        callbacks=[callback, saver_callback],
    )

    # save model with logist oututs (cut Softmax layer)
    model_logists = Model(model.input, model.layers[-2].output)
    if not os.path.isdir(str(args.models_path)):
        os.mkdir(str(args.models_path))
    model_logists.save(str(args.models_path / (config_name + '.h5')))
