# coding: utf-8
__author__ = 'ZFTurbo: https://www.drivendata.org/users/ZFTurbo/'

import numpy as np
import gzip
import pickle
import os
import sys
import glob
import time
import datetime
import pandas as pd
from sklearn.model_selection import KFold, train_test_split
from collections import Counter, defaultdict
from sklearn.metrics import matthews_corrcoef, roc_auc_score, accuracy_score
import random
import shutil
import operator
import platform
import json
import base64
import zlib
import gc
from constants import *

random.seed(2016)
np.random.seed(2016)


def save_in_file(arr, file_name):
    pickle.dump(arr, gzip.open(file_name, 'wb+', compresslevel=3), protocol=4)


def load_from_file(file_name):
    return pickle.load(gzip.open(file_name, 'rb'))


def save_in_file_fast(arr, file_name):
    pickle.dump(arr, open(file_name, 'wb'), protocol=4)


def load_from_file_fast(file_name):
    return pickle.load(open(file_name, 'rb'))


def get_date_string():
    return datetime.datetime.now().strftime("%Y-%m-%d-%H-%M")


def sort_dict_by_values(a, reverse=True):
    sorted_x = sorted(a.items(), key=operator.itemgetter(1), reverse=reverse)
    return sorted_x


def value_counts_for_list(lst):
    a = dict(Counter(lst))
    a = sort_dict_by_values(a, True)
    return a


def save_history_figure(history, path, columns=('acc', 'val_acc')):
    import matplotlib.pyplot as plt
    s = pd.DataFrame(history.history)
    plt.plot(s[list(columns)])
    plt.savefig(path)
    plt.close()


def get_model_memory_usage(batch_size, model):
    import numpy as np
    try:
        from keras import backend as K
    except:
        from tensorflow.keras import backend as K

    shapes_mem_count = 0
    internal_model_mem_count = 0
    for l in model.layers:
        layer_type = l.__class__.__name__
        if layer_type == 'Model':
            internal_model_mem_count += get_model_memory_usage(batch_size, l)
        single_layer_mem = 1
        out_shape = l.output_shape
        if type(out_shape) is list:
            out_shape = out_shape[0]
        for s in out_shape:
            if s is None:
                continue
            single_layer_mem *= s
        shapes_mem_count += single_layer_mem

    trainable_count = np.sum([K.count_params(p) for p in model.trainable_weights])
    non_trainable_count = np.sum([K.count_params(p) for p in model.non_trainable_weights])

    number_size = 4.0
    if K.floatx() == 'float16':
        number_size = 2.0
    if K.floatx() == 'float64':
        number_size = 8.0

    total_memory = number_size * (batch_size * shapes_mem_count + trainable_count + non_trainable_count)
    gbytes = np.round(total_memory / (1024.0 ** 3), 3) + internal_model_mem_count
    return gbytes


def normalize_array(cube, new_max, new_min):
    """Rescale an arrary linearly."""
    minimum, maximum = np.min(cube), np.max(cube)
    if maximum - minimum != 0:
        m = (new_max - new_min) / (maximum - minimum)
        b = new_min - m * minimum
        cube = m * cube + b
    return cube


def reduce_model(model_path):
    from kito import reduce_keras_model
    from keras.models import load_model

    m = load_model(model_path)
    m_red = reduce_keras_model(m)
    m_red.save(model_path[:-3] + '_reduced.h5')



def merge_models(shape_size, models_list, out_model_file, avg=True):
    from keras.models import load_model, save_model, Model
    from keras.layers import Input, Average

    if not os.path.isfile(out_model_file):
        models = []
        for m_path in models_list:
            m = load_model(m_path)
            models.append(m)

        inp = Input(shape_size)
        outs = []
        for i, m in enumerate(models):
            m.name += '_{}'.format(i)
            x = m(inp)
            outs.append(x)

        if avg is True:
            outs = Average()(outs)
        final_model = Model(inputs=inp, outputs=outs)
        print(final_model.summary())
        print('Single model memory: {} GB'.format(get_model_memory_usage(1, final_model)))
        save_model(final_model, out_model_file)
    else:
        final_model = load_model(out_model_file)

    return final_model


def get_classes():
    s = pd.read_csv(INPUT_PATH + 'train_labels.csv', nrows=2)
    cls = s.columns.values[1:]
    return cls


def convert_sequence_v1(s, length):
    out = np.zeros((length, 5), dtype=np.float32)
    s = np.array(list(s))
    out[:len(s), 0][s == 'C'] = 1
    out[:len(s), 1][s == 'A'] = 1
    out[:len(s), 2][s == 'T'] = 1
    out[:len(s), 3][s == 'G'] = 1
    out[:len(s), 4][s == 'N'] = 1
    return out


def convert_sequence(s, length):
    out = np.zeros((length, 5), dtype=np.float32)
    s = np.array(list(s))
    sl = len(s)
    start = (length - sl) // 2
    end = start + sl
    out[start:end, 0][s == 'C'] = 1
    out[start:end, 1][s == 'A'] = 1
    out[start:end, 2][s == 'T'] = 1
    out[start:end, 3][s == 'G'] = 1
    out[start:end, 4][s == 'N'] = 1
    return out


def accuracy_multi(real, pred, count=10):
    """
    if count == 1 default accuracy

    :param real:
    :param pred:
    :param count:
    :return:
    """
    r = np.argmax(real, axis=1)
    p = np.argsort(pred, axis=1)[:, -count:]
    score = 0.0
    for i in range(len(r)):
        if r[i] in p[i]:
            score += 1.0
    score /= len(r)
    return score


def invert_sequence(s):
    s = s.replace('T', '.')
    s = s.replace('A', ',')
    s = s.replace('.', 'A')
    s = s.replace(',', 'T')
    s = s.replace('C', '.')
    s = s.replace('G', ',')
    s = s.replace('.', 'G')
    s = s.replace(',', 'C')
    return s[::-1]


def softmax(x):
    """Compute softmax values for each sets of scores in x."""
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=0) # only difference


def normalize_submission(subm_path):
    CLASSES = get_classes()
    s = pd.read_csv(subm_path)
    matrix = s[CLASSES].values.astype(np.float64)
    if matrix.max() > 1000:
        matrix /= 1000
    for i in range(len(s)):
        print(i)
        matrix[i] = softmax(matrix[i])
    s[CLASSES] = matrix
    s.to_csv(subm_path[:-4] + '_norm.csv', index=False)


def cut_softmax_from_model(model):
    from keras.models import Model
    model_upd = Model(inputs=model.inputs, outputs=model.layers[-2].output)
    return model_upd


def get_simple_labels():
    s = pd.read_csv(INPUT_PATH + 'train_labels.csv')
    cls = s.columns.values[1:]
    s['target_name'] = s[cls].idxmax(axis=1)
    target = []
    cls = list(cls)
    for i in range(len(s)):
        val = cls.index(s['target_name'][i])
        target.append(val)
    s['target'] = target
    # s[['sequence_id', 'target_name', 'target']].to_csv(INPUT_PATH + 'train_labels_simple.csv', index=False)
    return s[['sequence_id', 'target_name', 'target']]


def get_id_to_class_dict():
    train = get_simple_labels()
    sequence_ids = train['sequence_id']
    targets = train['target']
    res = dict()
    for i in range(len(sequence_ids)):
        res[sequence_ids[i]] = targets[i]
    return res


def get_class_counts():
    res = dict()
    answs = get_id_to_class_dict()
    for el in answs:
        cls = answs[el]
        if cls not in res:
            res[cls] = 0
        res[cls] += 1
    # print(len(res))
    return res


def get_seq_lengths():
    train = pd.read_csv(INPUT_PATH + 'train_values.csv')
    test = pd.read_csv(INPUT_PATH + 'test_values.csv')
    res = dict()
    sequence_ids = list(train['sequence_id'].values) + list(test['sequence_id'].values)
    sequences = list(train['sequence'].values) + list(test['sequence'].values)

    for i in range(len(sequence_ids)):
        res[sequence_ids[i]] = len(sequences[i])
    return res
