import torch
import torch.nn as nn
import torch.nn.functional as F


class CNNText(nn.Module):
    def __init__(self, embed_dim=256, dropout=0.25, embed_num=100001, kernel_num=256, kernel_sizes=tuple(range(1, 8, 2)),
                 num_class=1314):
        super(CNNText, self).__init__()
        self.embed = nn.Embedding(embed_num, embed_dim)
        self.convs = nn.ModuleList([nn.Conv2d(1, kernel_num, kernel_size=(K, embed_dim)) for K in kernel_sizes])
        self.dropout = nn.Dropout(dropout)
        self.fc1 = nn.Linear(len(kernel_sizes) * kernel_num, num_class)

    def forward(self, x, rev_x):
        x = self.embed(x)  # (N, W, D)
        rev_x = self.embed(rev_x)
        x = x.unsqueeze(1)  # (N, Ci, W, D)
        rev_x = rev_x.unsqueeze(1)
        x = [F.relu(conv(x)).squeeze(3) for conv in self.convs]  # [(N, Co, W), ...]*len(Ks)
        rev_x = [F.relu(conv(rev_x)).squeeze(3) for conv in self.convs]
        x = [F.max_pool1d(i, i.size(2)).squeeze(2) for i in x]  # [(N, Co), ...]*len(Ks)
        rev_x = [F.max_pool1d(i, i.size(2)).squeeze(2) for i in rev_x]
        x = torch.cat(x, 1)
        rev_x = torch.cat(rev_x, 1)
        x = self.dropout(x)  # (N, len(Ks)*Co)
        rev_x = self.dropout(rev_x)
        logit = self.fc1(x)  # (N, C)
        logit_rev = self.fc1(rev_x)
        return logit+logit_rev


class JointModel(nn.Module):
    def __init__(self, seq_embed=CNNText, len_num=39, n_hiddens=(2048, 512), num_class=1314):
        super(JointModel, self).__init__()
        self.seq_embed = seq_embed(num_class=n_hiddens[0])
        self.num_proj = nn.Sequential(
            nn.Linear(len_num, 1024),
            nn.BatchNorm1d(1024),
            nn.ReLU(True),
            nn.Dropout(0.5),
            nn.Linear(1024, n_hiddens[1])
        )

        self.classifier = nn.Sequential(
            nn.Linear(sum(n_hiddens), 2048),
            nn.BatchNorm1d(2048),
            nn.ReLU(True),
            nn.Dropout(0.25),
            nn.Linear(2048, num_class)
        )

    def forward(self, seq, rev_seq, num_fea):
        embed_hidden = self.seq_embed(seq, rev_seq)
        num_hidden = self.num_proj(num_fea)
        joint_hidden = torch.cat([embed_hidden, num_hidden], 1)
        return self.classifier(joint_hidden)
