import torch
import os
from tqdm.auto import tqdm
import numpy as np
from torch.utils.data import DataLoader
from torch import nn
from data import Dataset
from transformers import AdamW, get_cosine_with_hard_restarts_schedule_with_warmup


def top10_acc(y_pred, y):
    top10 = np.argpartition(y_pred, -10, axis=1)[:, -10:]
    mask = top10 == y.reshape((y.size, 1))
    return mask.any(axis=1).mean()


class NNTrainer:
    def __init__(self, model_cls, max_len, loss_fn=torch.nn.CrossEntropyLoss(), opt=AdamW,
                 dset=Dataset, model_state_dict=None, model_path='model'):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.num_cudas = torch.cuda.device_count()
        self.model = model_cls.to(self.device)
        self.loss_fn = loss_fn
        self.dset = dset
        self.optimizer = opt(self.model.parameters(), lr=0.001)
        self.scheduler_func = get_cosine_with_hard_restarts_schedule_with_warmup
        self.model_path = model_path
        self.max_len = max_len

        if model_state_dict is not None:
            self.model.load_state_dict(model_state_dict)

        if self.num_cudas > 1:
            print(f"running on {self.num_cudas} GPUS")
            self.model = nn.DataParallel(self.model)

    def fit(self, train_set, test_set, num_epochs=10, early_stop=3, batch_size=128, shuffle=True,
            num_workers=8, verbose=True):

        optimizer = self.optimizer
        steps_per_epoch = int(np.ceil(train_set.shape[0]/batch_size))
        num_warmup_steps = steps_per_epoch
        num_training_steps = steps_per_epoch * num_epochs

        scheduler = self.scheduler_func(optimizer, num_warmup_steps=num_warmup_steps,
                                        num_training_steps=num_training_steps)

        val_scores = []

        CLIP = None
        dataloader = {
            'train': DataLoader(self.dset(train_set, max_len=self.max_len), batch_size=batch_size, shuffle=shuffle,
                                num_workers=num_workers),
            'val': DataLoader(self.dset(test_set, max_len=self.max_len), batch_size=batch_size, shuffle=False,
                              num_workers=num_workers)}

        for epoch in range(num_epochs):

            if verbose:
                print('Epoch {}/{}'.format(epoch, num_epochs))

            """Training phase"""
            self.model.train()
            preds, truths = [], []
            train_loss = 0
            for batch_data in tqdm(dataloader['train']) if verbose else dataloader['train']:
                seq, num_fea, target = batch_data['sequence'], batch_data['num_fea'], batch_data['target']
                rev_seq = batch_data['rev_seq'].to(self.device)
                seq, num_fea, target = seq.to(self.device), num_fea.to(self.device), target.to(self.device)

                output = self.model(seq, rev_seq, num_fea)
                loss = self.loss_fn(output, target)
                train_loss += loss.item()
                loss.backward()
                if CLIP is not None:
                    nn.utils.clip_grad_norm_(self.model.parameters(), CLIP)
                optimizer.step()
                scheduler.step()
                optimizer.zero_grad()
                preds.append(output.cpu().data.numpy())
                truths.append(target.cpu().data.numpy())
                del batch_data, seq, num_fea, target, rev_seq
                if self.num_cudas > 0:
                    torch.cuda.empty_cache()
            preds, truths = np.vstack(preds), np.hstack(truths)
            tr_top10 = top10_acc(preds, truths)
            if verbose:
                print(f"training top10 acc: {tr_top10}, loss: {train_loss / len(dataloader['train'])}")

            """Validating phase"""
            self.model.eval()
            tta_n = 1
            for tta_s in range(tta_n):
                val_loss = 0
                preds, truths = [], []
                for batch_data in tqdm(dataloader['val']):
                    seq, num_fea, target = batch_data['sequence'], batch_data['num_fea'], batch_data['target']
                    rev_seq = batch_data['rev_seq'].to(self.device)
                    seq, num_fea, target = seq.to(self.device), num_fea.to(self.device), target.to(self.device)

                    with torch.no_grad():
                        output = self.model(seq, rev_seq, num_fea)
                        loss = self.loss_fn(output, target)
                        val_loss += loss.item()

                    preds.append(output.cpu().data.numpy())
                    truths.append(target.cpu().data.numpy())
                    del batch_data, seq, num_fea, target, rev_seq
                    if self.num_cudas > 0:
                        torch.cuda.empty_cache()
                preds, truths = np.vstack(preds), np.hstack(truths)
                if tta_s == 0:
                    tta_preds = preds/tta_n
                else:
                    tta_preds += preds/tta_n
                va_top10 = top10_acc(tta_preds, truths)
                if verbose:
                    print(f"{tta_s} validation top10 acc: {va_top10}, loss: {val_loss/len(dataloader['val'])}")

            epoch_top10 = va_top10

            if epoch > 15:
                self.save_model_weights(epoch, score=epoch_top10)
            if (epoch == 0) or (epoch_top10 > max(val_scores)):
                best_epoch = epoch
                best_score = epoch_top10
                if verbose:
                    print(f"found a better acc at epoch {epoch}: {epoch_top10}")
            val_scores.append(epoch_top10)
            if (len(val_scores) > early_stop) and (max(val_scores) not in val_scores[-early_stop:]):
                if verbose:
                    print(f"early stop at {epoch}, with best epoch {best_epoch} : {best_score}")
                break
        return best_score

    def save_model_weights(self, epoch, score):
        torch.save(self.model.state_dict(), f"{self.model_path}/{score:.4f}_{epoch}.pt")

    def predict(self, data, batch_size=256, num_workers=8, topk=2):
        avg_preds = None
        saved_models = os.listdir(self.model_path)
        best_scores = sorted([float(x.split("_")[0]) for x in saved_models])[-topk:]
        best_models = [x for x in saved_models if float(x.split("_")[0]) in best_scores]

        dataloader = DataLoader(self.dset(data, max_len=self.max_len), batch_size=batch_size, num_workers=num_workers, shuffle=False)
        for m in best_models:
            print(f"predicting with model {m}")
            self.model.load_state_dict(torch.load(f"{self.model_path}/{m}"))
            self.model.eval()
            preds = []
            for batch_data in dataloader:
                seq, num_fea = batch_data['sequence'], batch_data['num_fea']
                rev_seq = batch_data['rev_seq'].to(self.device)
                seq, num_fea = seq.to(self.device), num_fea.to(self.device)

                with torch.no_grad():
                    output = self.model(seq, rev_seq, num_fea)
                    out = nn.Softmax(dim=1)(output)
                    preds.append(out.cpu().data.numpy())
            if avg_preds is None:
                avg_preds = np.vstack(preds)/len(best_models)
            else:
                avg_preds += np.vstack(preds)/len(best_models)

        return avg_preds
