import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from itertools import combinations, combinations_with_replacement, permutations
from sklearn.preprocessing import MinMaxScaler
from itertools import permutations
from tqdm import tqdm
from sklearn.preprocessing import PolynomialFeatures
from sklearn.mixture import GaussianMixture
import Bio
from Bio.Seq import Seq
import re
import gc
import os

import warnings
warnings.filterwarnings("ignore")
plt.rc('font', size=14)
pd.set_option('display.max_columns', 500)


def get_ngram_features(sequence, subsequences):
    """Generates counts for each subsequence.

    Args:
        data (DataFrame): The data you want to create features from. Must include a "sequence" column.
        subsequences (list): A list of subsequences to count.

    Returns:
        DataFrame: A DataFrame with one column for each subsequence.
    """
    features = pd.DataFrame(index=sequence.index)
    for subseq in tqdm(subsequences):
        features[subseq] = sequence.str.count(subseq).astype('float32')
    return features


# Load data
y_train = pd.read_csv('../../data/raw/train_labels.csv').drop('sequence_id', axis=1)
X_train = pd.read_csv('../../data/raw/train_values.csv').drop('sequence_id', axis=1)
X_test = pd.read_csv('../../data/raw/test_values.csv').drop('sequence_id', axis=1)
test_sequence_id = pd.read_csv('../../data/raw/test_values.csv')['sequence_id']

# convert types
X_train[X_train.columns.drop('sequence')] = X_train[X_train.columns.drop('sequence')].astype(int)
X_test[X_test.columns.drop('sequence')] = X_test[X_test.columns.drop('sequence')].astype(int)


one_hot_columns = X_train.columns.drop('sequence').tolist()


# Feature engineering.
# Sequence length
X_train['sequence_len'] = X_train.sequence.apply(len)
X_test['sequence_len'] = X_test.sequence.apply(len)
X_train['sequence_len_log'] = np.log(X_train['sequence_len'])
X_test['sequence_len_log'] = np.log(X_test['sequence_len'])

# Clusters based on bimodal distribution
np.random.seed(0)
EM = GaussianMixture(n_components=2, random_state=0)
EM.fit(X_train['sequence_len_log'].to_frame())
X_train['EM_cluster'] = EM.predict(X_train['sequence_len_log'].to_frame())
X_test['EM_cluster'] = EM.predict(X_test['sequence_len_log'].to_frame())
# save probabilities
train_proba = EM.predict_proba(X_train['sequence_len_log'].to_frame())
test_proba = EM.predict_proba(X_test['sequence_len_log'].to_frame())
if train_proba.shape[1] > 2:
    for i in range(train_proba.shape[1]):
        X_train[f'EM_cluster_{i}'] = train_proba[:, i]
        X_test[f'EM_cluster_{i}'] = test_proba[:, i]
else:
    X_train[f'EM_cluster_1_prob'] = train_proba[:, 1]
    X_test[f'EM_cluster_1_prob'] = test_proba[:, 1]

# Sequence based features.
bases = ['A', 'C', 'G', 'T', 'N']
for base in bases:
    fig, axes = plt.subplots(1, 2, figsize=(16, 3))
    train_freq = (X_train.sequence.str.count(base) / X_train['sequence_len'])
    test_freq = (X_test.sequence.str.count(base) / X_test['sequence_len'])
    axes[0].set_title(base + ' train mean:' + str(round(train_freq.mean(), 5)))
    axes[1].set_title(base + ' test mean:' + str(round(test_freq.mean(), 5)))
    train_freq.hist(bins=200, ax=axes[0])
    test_freq.hist(bins=200, ax=axes[1])

# Base count and frequency features.
bases = ['C', 'G', 'N']
for base in bases:
    # counts
    X_train[base+'_c'] = X_train.sequence.str.count(base)
    X_test[base+'_c'] = X_test.sequence.str.count(base)
    # frequencies
    X_train[base+'_f'] = X_train[base+'_c'] / X_train['sequence_len']
    X_test[base+'_f'] = X_test[base+'_c'] / X_test['sequence_len']

# GC Content
X_train['GC_c'] = X_train['G_c'] + X_train['C_c']
X_test['GC_c'] = X_test['G_c'] + X_test['C_c']

# GC frequencies
X_train['GC_f'] = X_train['G_f'] + X_train['C_f']
X_test['GC_f'] = X_train['G_f'] + X_train['C_f']

# drop G's and C's
X_train.drop(['G_f', 'G_c', 'C_f', 'C_c'], axis=1, inplace=True)
X_test.drop(['G_f', 'G_c', 'C_f', 'C_c'], axis=1, inplace=True)

# Polynoms of one-hot encoded features
poly = PolynomialFeatures(interaction_only=True)

train_poly = pd.DataFrame(poly.fit_transform(X_train[one_hot_columns])).add_prefix('poly_').astype('int8')
test_poly = pd.DataFrame(poly.transform(X_test[one_hot_columns])).add_prefix('poly_').astype('int8')

# drop null columns
cols_to_drop = train_poly.columns[train_poly.nunique() == 1]
train_poly = train_poly.drop(cols_to_drop, axis=1)
test_poly = test_poly.drop(cols_to_drop, axis=1)

# Drop identical columns
train_poly = train_poly.T.drop_duplicates().T
test_poly = test_poly[train_poly.columns]

X_train = X_train.join(train_poly)
X_test = X_test.join(test_poly)

# Drop identical columns
X_train = X_train.T.drop_duplicates().T
X_test = X_test[X_train.columns]


# Additional strings
orf_expression = r'ATG(?:(?!TAA|TAG|TGA)...)*(?:TAA|TAG|TGA)'

# reverse complement
X_train['sequence_rc'] = X_train['sequence'].apply(lambda x: str(Seq(x).reverse_complement()))
X_test['sequence_rc'] = X_test['sequence'].apply(lambda x: str(Seq(x).reverse_complement()))

# Only open reading frame (ORF), separated using space character
X_train['orf_sequence'] = X_train['sequence'].apply(lambda x: ' '.join(re.findall(orf_expression, x)))
X_test['orf_sequence'] = X_test['sequence'].apply(lambda x: ' '.join(re.findall(orf_expression, x)))
X_train['orf_sequence_rc'] = X_train['sequence_rc'].apply(lambda x: ' '.join(re.findall(orf_expression, x)))
X_test['orf_sequence_rc'] = X_test['sequence_rc'].apply(lambda x: ' '.join(re.findall(orf_expression, x)))

seq_features = ['sequence', 'sequence_rc', 'orf_sequence', 'orf_sequence_rc',]


# Minmax scale
# features for scaling
poly_cols = X_test.columns[X_test.columns.str.startswith('poly_')].tolist()
scaling_cols = X_test.columns.drop(seq_features + one_hot_columns + poly_cols).tolist()

scaler = MinMaxScaler()
scaler.fit(X_train[scaling_cols])
X_train[scaling_cols] = scaler.transform(X_train[scaling_cols]).astype('float32')
X_test[scaling_cols] = scaler.transform(X_test[scaling_cols]).astype('float32')
del scaler


# Convert types
X_train[one_hot_columns] = X_train[one_hot_columns].astype('uint8')
X_test[one_hot_columns] = X_test[one_hot_columns].astype('uint8')
y_train = y_train.astype('uint8')


# Kmer based features
bases = ['A', 'C', 'G', 'T']
k = 6

# get all possible kmer combinations
subsequences_full = [[''.join(p) for p in permutations(''.join(c), r=k)] for c in combinations_with_replacement(bases, r=k)]
subsequences = []
for sub in subsequences_full:
    subsequences.extend(sub)
subsequences = sorted(list(set(subsequences)))


# Regular k-mer count for a string and its reverse complement
# Get n-gram features
train_kmers = get_ngram_features(X_train['sequence'], subsequences)
test_kmers = get_ngram_features(X_test['sequence'], subsequences)

# get reverse complements of kmer_counts
#(it's faster to reverse existing column names rather than counting reversed string)
train_kmers_rc = train_kmers.copy()
test_kmers_rc = test_kmers.copy()
train_kmers_rc.columns = test_kmers_rc.columns = list(map(lambda x: str(Seq(x).reverse_complement()), 
                                                          train_kmers.columns.tolist()))
# get sum
train_kmers_rc_s = train_kmers + train_kmers_rc
test_kmers_rc_s = test_kmers + test_kmers_rc

# Drop identical columns
train_kmers_rc_s = train_kmers_rc_s.T.drop_duplicates().T
test_kmers_rc_s = test_kmers_rc_s[train_kmers_rc_s.columns]

# Scaling, works part by part to avoid memory problems
print('scaling kmers...')
for part in range(0, train_kmers_rc_s.shape[1], 2048):
    scaler = MinMaxScaler()
    scaler.fit(train_kmers_rc_s.iloc[:, part: part + 2048])
    train_kmers_rc_s.iloc[:, part: part + 2048] = scaler.transform(train_kmers_rc_s.iloc[:, part: part+2048]).astype('float32')
    test_kmers_rc_s.iloc[:, part: part + 2048] = scaler.transform(test_kmers_rc_s.iloc[:, part: part+2048]).astype('float32')
print('done!')

# join kmers
X_train = X_train.join(train_kmers_rc_s.add_prefix('kmers_'))
X_test = X_test.join(test_kmers_rc_s.add_prefix('kmers_'))

del train_kmers, test_kmers, train_kmers_rc, test_kmers_rc, train_kmers_rc_s, test_kmers_rc_s

# Open Reading Frames
# Get n-gram features regualar + rev comp
orf_train_kmers = get_ngram_features(X_train['orf_sequence'] + ' ' + X_train['orf_sequence_rc'], subsequences)
orf_test_kmers = get_ngram_features(X_test['orf_sequence'] + ' ' + X_test['orf_sequence_rc'], subsequences)

# Drop identical columns
orf_train_kmers = orf_train_kmers.T.drop_duplicates().T
orf_test_kmers = orf_test_kmers[orf_train_kmers.columns]

# Scaling, works part by part to avoid memory problems
print('scaling kmers...')
for part in range(0, orf_train_kmers.shape[1], 2048):
    scaler = MinMaxScaler()
    scaler.fit(orf_train_kmers.iloc[:, part: part + 2048])
    orf_train_kmers.iloc[:, part: part + 2048] = scaler.transform(orf_train_kmers.iloc[:, part: part + 2048]).astype('float32')
    orf_test_kmers.iloc[:, part: part + 2048] = scaler.transform(orf_test_kmers.iloc[:, part: part + 2048]).astype('float32')
print('done!')

# join orf kmers
X_train = X_train.join(orf_train_kmers.add_prefix('orf_'))
X_test = X_test.join(orf_test_kmers.add_prefix('orf_'))

# drop sequence features
X_train = X_train.drop(seq_features, axis=1)
X_test = X_test.drop(seq_features, axis=1)

# Save
X_train.to_pickle('../../data/processed/gk_X_train.pk')
X_test.to_pickle('../../data/processed/gk_X_test.pk')
