import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from datetime import datetime
from sklearn.model_selection import StratifiedKFold
import tensorflow as tf
import os
import gc
from tqdm import tqdm
import warnings
warnings.filterwarnings("ignore")
plt.rc('font', size=14)
pd.set_option('display.max_columns', 500)

 
def top10_accuracy_score(pred, y):
    top10_idx = np.argpartition(pred, -10, axis=1)[:, -10:]
    mask = top10_idx == y.reshape((y.size, 1))
    top_10_accuracy = mask.any(axis=1).mean()
    return top_10_accuracy


def top10_accuracy_score_per_class(pred, y):
    top10_idx = np.argpartition(pred, -10, axis=1)[:, -10:]
    mask = top10_idx == y.reshape((y.size, 1))
    is_correct = mask.any(axis=1)
    
    # concat class and true/false value
    top_10_per_class = np.concatenate([y.reshape((y.size, 1)), is_correct.reshape((is_correct.size, 1))], 
                                      axis=1)
    
    # group by class, get average per class
    top_10_per_class = pd.DataFrame(top_10_per_class, columns=['class', 'is_correct']).groupby('class').mean()
    
    # average the averages
    top_10_accuracy_reg = top_10_per_class.mean().values[0]
    return top_10_accuracy_reg


def get_nn_model(n_features, n_layers, layer_size, learning_rate, dropout_rate=0, num_classes=None):
    metric = tf.keras.metrics.TopKCategoricalAccuracy(k=10, name="top_k_categorical_accuracy", dtype=None)
    optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)
    loss = tf.keras.losses.CategoricalCrossentropy(from_logits=False)

    input_other = tf.keras.layers.Input(shape=(n_features,))
    
    for i in range(n_layers):
        if i == 0:
            x = tf.keras.layers.Dense(layer_size, activation='relu')(input_other)
        else:
            x = tf.keras.layers.Dense(layer_size, activation='relu')(x)
        if dropout_rate > 0:
            x = tf.keras.layers.Dropout(dropout_rate, seed=0,)(x)
    
    output = tf.keras.layers.Dense(num_classes)(x)
        
    output = tf.keras.layers.Activation('softmax')(output)
    model = tf.keras.Model(inputs=input_other, outputs=output)

    model.compile(optimizer=optimizer, loss=loss, metrics=[metric])
    return model


def predict_nn_cv(model_name, model_class, model_params, X_train, y_train, X_test=None,
                  batch_size=512, early_stopping_rounds=20, cv=5, verbose=0):
    
    train_pred = pd.DataFrame(np.zeros(y_train.shape))
    
    if X_test is not None:
        test_pred = 0   
        
    weights = []

    for x, ind in enumerate(cv):
        start = datetime.now()
        
        train_index, valid_index = ind
        
        # Get model instance
        model = model_class(**model_params)
        
        model = tf.keras.models.load_model(f'{MODEL_FOLDER}/{model_name}_fold_{x}.h5')
        
        # Predict validation.
        valid_pred = model.predict(X_train.iloc[valid_index].values)
        train_pred.iloc[valid_index] = valid_pred.astype('float16')
        
        # Predict test.
        if X_test is not None:
            fold_test_pred = model.predict(X_test.values)
            test_pred += fold_test_pred
        
        # Get top10_accuracy score for a fold.
        fold_score = top10_accuracy_score(valid_pred, y_train.idxmax(axis=1).iloc[valid_index].values)
        print(f'fold {x}, top10_accuracy:', round(fold_score, 5), end=', ')
        print(f'time:', str(datetime.now() - start)[:7])
        
        # Clear memory
        gc.collect()
 
    if X_test is not None:
        # Average test predictions
        test_pred = pd.DataFrame(test_pred) / N_SPLITS
        return train_pred, test_pred
    else:
        return train_pred, pd.DataFrame([])


# Settings.
N_SPLITS = 5  # kfold settings
SHUFFLE = True  # kfold settings

N_SEEDS = 5
BATCH_SIZE = 1024
EARLY_STOPPING_ROUNDS = 30
VERBOSE = 0
MODEL_FOLDER = '../../models'


# Load data
y_train = pd.read_csv('../../data/raw/train_labels.csv').drop('sequence_id', axis=1)
X_train = pd.read_pickle('../../data/processed/gk_X_train.pk').astype('float32')
X_test = pd.read_pickle('../../data/processed/gk_X_test.pk').astype('float32')

# save column names
y_train_columns = y_train.columns.tolist()
# rename columns
y_train.columns = list(range(y_train.shape[1]))


# Predicting
train_preds, test_preds = {}, {}

np.random.seed(0)
skf = StratifiedKFold(n_splits=N_SPLITS, random_state=0, shuffle=SHUFFLE)
folds = list(skf.split(X_train, y_train.idxmax(axis=1)))

# Predicting only 6mers
orf_columns = X_train.columns[X_train.columns.str.startswith('orf')].tolist()

name = 'gk_6mers'
model_params = {
    'n_features': X_train.drop(orf_columns, axis=1).shape[1], 'n_layers': 1, 'layer_size': 2048,               
    'learning_rate': 0.0005, 'dropout_rate': 0.7, 'num_classes': y_train.shape[1],
}

for i in range(N_SEEDS):
    print(name + f'_seed{i}')
    start = datetime.now()
    
    train_pred, test_pred = predict_nn_cv(f'{name}_seed{i}', get_nn_model, model_params,
                                          X_train.drop(orf_columns, axis=1),  # drop orf for this model
                                          y_train,
                                          X_test.drop(orf_columns, axis=1),  # drop orf for this model
                                          batch_size=BATCH_SIZE, cv=folds, verbose=VERBOSE,
                                          early_stopping_rounds=EARLY_STOPPING_ROUNDS)
    
    # get score
    score = top10_accuracy_score(train_pred.values, y_train.idxmax(axis=1).values)
    print('out-of-fold top10_accuracy:', round(score, 5), ', time:', str(datetime.now() - start)[:7])
    train_preds[name + str(i)] = train_pred
    test_preds[name + str(i)] = test_pred


# Predicting 6mers + 6mers from open reading frames
name = 'gk_6mers_orf6mers'
model_params = {
    'n_features': X_train.shape[1], 'n_layers': 1, 'layer_size': 1024,               
    'learning_rate': 0.0005, 'dropout_rate': 0.7, 'num_classes': y_train.shape[1],
}

for i in range(N_SEEDS):
    print(name + f'_seed{i}')
    start = datetime.now()
    
    train_pred, test_pred = predict_nn_cv(f'{name}_seed{i}', get_nn_model, model_params,
                                          X_train, y_train, X_test,
                                          batch_size=BATCH_SIZE, cv=folds, verbose=VERBOSE,
                                          early_stopping_rounds=EARLY_STOPPING_ROUNDS)
    
    # get score
    score = top10_accuracy_score(train_pred.values, y_train.idxmax(axis=1).values)
    print('out-of-fold top10_accuracy:', round(score, 5), ', time:', str(datetime.now() - start)[:7])
    train_preds[name + str(i)] = train_pred
    test_preds[name + str(i)] = test_pred

    
# average results per model
print('\n')
for group in ['gk_6mers', 'gk_6mers_orf6mers']:
    subgroup = [group + str(n) for n in range(N_SEEDS)]

    train_mean = 0
    test_mean = 0
    for i, pred in enumerate(subgroup):
        train_mean += train_preds[pred]
        test_mean += test_preds[pred]

    train_preds[f'{group}_average'] = train_mean / len(subgroup)
    test_preds[f'{group}_average'] = test_mean / len(subgroup)

    # calculace proper log loss score
    avg_score = top10_accuracy_score(train_preds[f'{group}_average'].values, y_train.idxmax(axis=1).values)
    print(f'{group} top10_accuracy:', round(avg_score, 5))


# weighted average based on top10_accuracy_score_per_class metric
avg_preds = ['gk_6mers_average', 'gk_6mers_orf6mers_average']

best_score, best_alpha, best_pred = 0, 0, train_preds[avg_preds[0]]
         
for alpha in tqdm(np.arange(0, 1.01, 0.01)):
    pred = alpha * train_preds[avg_preds[0]] + (1 - alpha) * train_preds[avg_preds[1]]
    score = top10_accuracy_score_per_class(pred.values, y_train.idxmax(axis=1).values)
    if score > best_score:
        best_score, best_alpha, best_pred = score, alpha, pred
    
print('\nweighted average, best alpha', best_alpha)
train_preds['gk_weighted_average'] = best_pred
test_preds['gk_weighted_average'] = best_alpha * test_preds[avg_preds[0]] + (1 - best_alpha) * test_preds[avg_preds[1]]


# final score
avg_score = top10_accuracy_score(train_preds['gk_weighted_average'].values, y_train.idxmax(axis=1).values)
avg_score_reg = top10_accuracy_score_per_class(train_preds['gk_weighted_average'].values, 
                                               y_train.idxmax(axis=1).values)
score_0 = top10_accuracy_score(train_preds['gk_weighted_average'][X_train['EM_cluster']==0].values,
                                           y_train[X_train['EM_cluster']==0].idxmax(axis=1).values)
score_1 = top10_accuracy_score(train_preds['gk_weighted_average'][X_train['EM_cluster']==1].values, 
                                           y_train[X_train['EM_cluster']==1].idxmax(axis=1).values)

print('\ngk_weighted_average')
print('top10_accuracy:', round(avg_score, 5))
print('top10_accuracy per class:', round(avg_score_reg, 5))
print('cluster 0 top10_accuracy:', round(score_0, 5))
print('cluster 1 top10_accuracy:', round(score_1, 5))

# proper columns names
train_preds['gk_weighted_average'].columns = y_train_columns
test_preds['gk_weighted_average'].columns = y_train_columns

# pickle predictions
# train_preds['gk_weighted_average'].to_pickle(f'../models/gk_train_preds.pk')
# test_preds['gk_weighted_average'].to_pickle(f'../models/gk_test_preds.pk')

# save predictions as csv
test_sequence_id = pd.read_csv('../../data/raw/test_values.csv')['sequence_id']
submission = test_sequence_id.to_frame().join(test_preds['gk_weighted_average'].astype('float32'))
submission.to_csv('../../models/gk_test_pred.csv', index=False)

train_sequence_id = pd.read_csv('../../data/raw/train_values.csv')['sequence_id']
train_submission = train_sequence_id.to_frame().join(pd.DataFrame(train_preds['gk_weighted_average']).astype('float32'))
train_submission.to_csv('../../models/gk_train_pred.csv', index=False)
